import sys, os, time
import numpy as np
import scipy as sci
import scipy.stats as ss
import scipy.sparse.linalg as slin
import copy
from .mytools.MinTree import MinTree
from scipy.sparse import coo_matrix, csr_matrix, lil_matrix
from .mytools.ioutil import loadedge2sm
from .edgepropertyAnalysis import MultiEedgePropBiGraph
import math
from .._model import DMmodel
from spartan.util.basicutil import param_default
from spartan.backend import STensor

def score_level_objects( objscores, p=0.90):
    '''implement with Perato distribution, given significant value
    '''
    sortscores = np.array(sorted(objscores))
    sortobjs = np.argsort(objscores)
    alpha = 0.9
    tail_fir_score = np.percentile(sortscores, [alpha*100])[0]
    if tail_fir_score == 0:
        'remove 0 if the number of percentile 90% is 0'
        firindex = np.argwhere(sortscores > 0)[0]
        sortscores = sortscores[firindex[0]:]
        sortobjs = sortobjs[firindex[0]:]
    'fit generalized pareto distribution using 10% upper tail data'
    tailidx = int(alpha * len(sortscores))
    tailscores = sortscores[tailidx:]
    tailobjs = sortobjs[tailidx:]

    shape, pos, scale = ss.pareto.fit(tailscores)
    cdfs = ss.pareto.cdf(tailscores, shape, pos, scale)

    levelidxs = np.argwhere(cdfs >= p)
    levelobjs = tailobjs[levelidxs].T[0]
    return levelobjs

def score_heristic_level_objects( objscores ):
    '''todo: implement with Perato distribution, given significant value
    '''
    sortscores = sorted(objscores, reverse=True)
    sortobjs = np.argsort(objscores)[::-1]
    diffscores = - np.diff(sortscores)
    levelid = np.argmax(diffscores)
    levelobjs = sortobjs[ : levelid+1]
    return levelobjs

def nonzero_objects( objscores ):
    objects = np.where( objscores > 0 )[0]
    return objects

class Ptype(object):
    freq =0
    ts = 1
    rate=2
    @staticmethod
    def ptype2str(p):
        if p == Ptype.freq:
            return 'freq'
        if p == Ptype.ts:
            return 'ts'
        if p == Ptype.rate:
            return 'rate'
    @staticmethod
    def ptypes2str(ptypes):
        strs=[]
        if Ptype.freq in ptypes:
            strs.append(Ptype.ptype2str(Ptype.freq))
        if Ptype.ts in ptypes:
            strs.append(Ptype.ptype2str(Ptype.ts))
        if Ptype.rate in ptypes:
            strs.append(Ptype.ptype2str(Ptype.rate))
        pstr = '-'.join(strs)
        return pstr

class HoloScopeOpt:
    def __init__(self, graphmat, qfun='exp', b=32,
                 aggmethod='sum', sdrop=True, mbd=0.5, sdropscale='linear',
                 tsprop=None, tunit='s', rateprop=None):
        'how many times of a user rates costumers if he get the cost balance'
        self.coe = 0
        'the larger expbase can give a heavy penalty to the power-law curve'
        self.expbase = b
        self.scale = qfun
        self.b = b
        self.aggmethod=aggmethod
        self.suspbd = 0.0 #susp < suspbd will assign to zero
        self.priordropslop=sdrop

        self.graph=graphmat.tocoo()
        self.graphr = self.graph.tocsr()
        self.graphc = self.graph.tocsc()
        self.matricizetenor=None
        self.nU, self.nV=graphmat.shape
        self.indegrees = graphmat.sum(0).getA1()
        self.e0 = math.log(graphmat.sum(), self.nU) #logrithm of edges
        print('matrix size: {} x {}\t#edges: {}'.format(self.nU, self.nV,
                                                          self.indegrees.sum()))

        # tunit is only used for files input
        self.tsprop, self.rateprop, self.tunit = tsprop, rateprop, tunit
        self.tspim, self.ratepim = None, None

        'field for multiple property graph'
        if tsprop is not None or rateprop is not None:
            if self.priordropslop:
                self.orggraph = self.graphr.copy()
            else:
                self.orggraph = self.graphr
        if tsprop is not None:
            self.mbd = mbd #multiburst bound
            self.tspim = MultiEedgePropBiGraph(self.orggraph)
            """
            since the data is cut by the end of time, so we need to see
            whether there is enough time twait from end of retweet to end of the
            whole data to judge if it is a sudden drop or cut by the end of time.
            twaits:
            """
            if isinstance(tsprop, str) and os.path.isfile(tsprop):
                self.tspim.load_from_edgeproperty(tsprop, mtype=coo_matrix,
                        dtype=np.int64)
                twaits = {'s':12*3600, 'h':24, 'd':30, None:0}
                twait = twaits[tunit]
            elif isinstance(tsprop, STensor):
                self.tspim.trans_array_to_edgeproperty(tsprop,
                        mtype=coo_matrix, dtype=np.int64)
                twait = 12
            else:
                raise Exception('Error: incorrect time stamp property')
            self.tspim.setup_ts4all_sinks(twait)
            if self.priordropslop:
                'slops weighted with max burst value'
                self.weightWithDropslop(weighted=True, scale=sdropscale)
        else:
            self.priordropslop = False #no input of time attribute
        if rateprop is not None:
            self.ratepim = MultiEedgePropBiGraph(self.orggraph)
            if isinstance(rateprop, str) and os.path.isfile(rateprop):
                self.ratepim.load_from_edgeproperty(rateprop, mtype=coo_matrix, dtype=float)
            elif isinstance(rateprop, STensor):
                self.ratepim.trans_array_to_edgeproperty(rateprop,
                        mtype=coo_matrix, dtype=float)
            else:
                raise Exception('Error: incorrect rate property')
            self.ratepim.setup_rate4all_sinks()

        'weighed with idf prior from Fraudar'
        #self.weightWithIDFprior()
        'if weighted the matrix the windegrees is not equal to indegrees'
        self.windegrees = self.graphc.sum(0).getA1()
        self.woutdegrees = self.graphr.sum(1).getA1()

        self.A = np.array([]) #binary array
        self.fbs = np.zeros(graphmat.shape[1], dtype=np.int) #frequency of bs in B
        '\frac_{ f_A{(bi)} }{ f_U{(bi)}}'
        self.bsusps = np.array([]) # the suspicious scores of products given A
        self.vx = 0 # current objective value
        self.vxs = [] #record all the vxs of optimizing iterations
        self.Y= np.array([])
        self.yfbs = np.array([])
        self.ybsusps = np.array([])
        'current is the best'
        self.bestvx = self.vx
        self.bestA = np.array([])
        self.bestfbs = np.array([])
        self.bestbsusps = np.array([])


    def weightWithDropslop(self, weighted, scale):
        'weight the adjacency matrix with the sudden drop of ts for each col'
        if weighted:
            colWeights = np.multiply(self.tspim.dropslops, self.tspim.dropfalls)
        else:
            colWeights = self.tspim.dropslops
        if scale == 'logistic':
            from scipy.stats import logistic
            from sklearn import preprocessing
            'zero mean scale'
            colWeights = preprocessing.scale(colWeights)
            colWeights = logistic.cdf(colWeights)
        elif scale == 'linear':
            from sklearn import preprocessing
            #add a base of suspecious for each edge
            colWeights = preprocessing.minmax_scale(colWeights) +1
        elif scale == 'plusone':
            colWeights += 1
        elif scale == 'log1p':
            colWeights = np.log1p(colWeights) + 1
        else:
            print('[Warning] no scale for the prior weight')

        n = self.nV
        colDiag = lil_matrix((n, n))
        colDiag.setdiag(colWeights)
        self.graphr = self.graphr * colDiag.tocsr()
        self.graph = self.graphr.tocoo(copy=False)
        self.graphc = self.graph.tocsc(copy=False)
        print("finished computing weight matrix")

    def weightWithIDFprior(self):
        print('weightd with IDF prior')
        colWeights = 1.0/np.log(self.indegrees + 5)
        n = self.nV
        colDiag = lil_matrix((n, n))
        colDiag.setdiag(colWeights)
        self.graphr = self.graphr * colDiag.tocsr()
        self.graph = self.graphr.tocoo(copy=False)
        self.graphc = self.graph.tocsc(copy=False)
        return

    'new objective with no f_A(v)/|A|'
    def maxobjfunc(self, A, fbs, bsusps=None):
        nu = 0.0
        de = 0.0
        numA = np.sum(A)
        de = numA + bsusps.sum() #math.sqrt(numA*bsusps.sum())#similar
        if numA == 0:
            return 0
        if bsusps is not None:
            nu = np.dot(fbs, bsusps)
        else:
            nu = fbs.sum()
        res = nu/np.float64( de )
        return res

    def aggregationMultiProp(self, mbs, method='sum'):
        if method == 'rank':
            from scipy.stats import rankdata
        rankmethod = 'average'
        k=60 #for rank fusion
        values = list(mbs.values())
        if len(mbs) == 1:
            val = values[0]
            if method == 'rank':
                rb = rankdata(-np.array(val), method=rankmethod)
                return np.reciprocal(rb+k) * k
            else:
                return val
        if method == 'sum':
            'this is the joint probability of exp form of prob'
            bsusps = values[0]
            for v in values[1:]:
                bsusps += v
        elif method == 'rank':
            'rank fusion'
            arrbsusps = []
            for val in values:
                rb = rankdata(-np.array(val), method=rankmethod)
                arrbsusps.append(np.reciprocal(rb+k))
            bsusps = np.array(arrbsusps).sum(0) * k
        else:
            print('[Error] Invalid method {}\n'.format(method))
        return bsusps

    #@profile
    def evalsusp4ts(self, suspusers, multiburstbd = 0.5, weighted=True):
        'the id of suspusers consistently starts from 0 no matter the source'
        incnt, inratio = self.tspim.suspburstinvolv(multiburstbd, weighted,
                                                    delta=True)
        suspts=inratio
        return suspts

    #@profile
    def evalsusp4rate(self, suspusers, neutral=False, scale='max'):
        susprates = self.ratepim.suspratedivergence(neutral, delta=True)
        if scale == 'max':
            if self.ratepim.maxratediv > 0:
                nsusprates = susprates/self.ratepim.maxratediv
            else:
                nsusprates = susprates
        elif scale=='minmax':
            #need a copy, and do not change susprates' value for delta
            from sklearn import preprocessing
            nsusprates = preprocessing.minmax_scale(susprates, copy=True)
        else:
            #no scale
            nsusprates = susprates
        return nsusprates

    'sink suspicious with qfunc, no f_A(v)/|A|'
    def prodsuspicious(self, fbs, A=None, scale='exp', ptype=[Ptype.freq]):
        multibsusps={}
        if Ptype.freq in ptype:
            posids = self.windegrees>0
            bs = np.zeros(self.nV)
            bs[posids] = np.divide(fbs[posids], self.windegrees[posids].astype(np.float64))
            multibsusps[Ptype.freq] = bs
        if Ptype.ts in ptype:
            suspusers = A.nonzero()[0]
            bs = self.evalsusp4ts(suspusers, multiburstbd=self.mbd)
            multibsusps[Ptype.ts] = bs
        if Ptype.rate in ptype:
            suspusers = A.nonzero()[0]
            bs = self.evalsusp4rate(suspusers)
            multibsusps[Ptype.rate] = bs
        bsusps = self.aggregationMultiProp(multibsusps, self.aggmethod)
        bsusps = self.qfunc(bsusps, fbs=fbs, scale=scale,
                numratios=len(multibsusps))
        return bsusps

    def initpimsuspects(self, suspusers, ptype):
        if Ptype.ts in ptype:
            self.tspim.setupsuspects(suspusers)
            temp1, temp2 = self.tspim.suspburstinvolv(multiburstbd=0.5, weighted=True,
                                       delta=False)
        if Ptype.rate in ptype:
            self.ratepim.setupsuspects(suspusers)
            tmp = self.ratepim.suspratedivergence(neutral=False,
                                            delta=False)
        return

    def start(self, A0, ptype=[Ptype.ts]):
        self.A = A0
        users = A0.nonzero()[0]
        self.ptype=ptype # the property type that the postiorer uses
        self.fbs = self.graphr[users].sum(0).getA1()
        self.fbs = self.fbs.astype(np.float64, copy=False)
        'initially set up currrent suspects'
        self.initpimsuspects(users, ptype=ptype)
        self.bsusps = self.prodsuspicious(self.fbs, self.A, ptype=ptype)
        self.vx = self.maxobjfunc(self.A, self.fbs, self.bsusps)
        self.vxs.append(self.vx)
        "current is the best"
        self.bestA = np.array(self.A)
        self.bestvx = self.vx
        self.bestfbs = np.array(self.fbs)
        self.bestbsusps = np.array(self.bsusps)

    def candidatefbs(self, z):
        'increase or decrease'
        coef = 1 if self.A[z] == 0 else -1
        bz = self.graphr[z]
        candfbs = (coef*bz + self.fbs).getA1()
        return candfbs

    #@profile
    def greedyshaving(self):
        '''greedy algorithm'''
        maxint = np.iinfo(np.int64).max//2
        delscores = np.array([maxint]*self.nU)
        delcands = self.A.nonzero()[0]
        deluserCredit = self.graphr[delcands,:].dot(self.bsusps)
        delscores[delcands] = deluserCredit
        print('set up the greedy min tree')
        MT = MinTree(delscores)
        i=0
        sizeA = np.sum(self.A)
        sizeA0 = sizeA
        setA = set(self.A.nonzero()[0])
        while len(setA) > 0:
            z, nextdelta = MT.getMin()
            setY = setA - {z}
            Y = copy.copy(self.A) # A is X
            Y[z] = 1-Y[z]
            self.Y=Y
            self.yfbs = self.candidatefbs(z)
            Ylist = Y.nonzero()[0]
            self.setdeltapimsusp(z, Ylist, add=False)
            self.ybsusps = self.prodsuspicious(self.yfbs, self.Y,
                                               ptype=self.ptype)
            vy = self.maxobjfunc(self.Y, self.yfbs, self.ybsusps)
            'chose next if next if the best'
            if vy > self.bestvx:
                self.bestA = np.array(self.Y)
                self.bestfbs = self.yfbs
                self.bestbsusps = self.ybsusps
                self.bestvx = vy
            MT.changeVal(z, maxint) #make the min to the largest for deletion
            prodchange = self.ybsusps - self.bsusps
            effectprod = prodchange.nonzero()[0]
            if len(effectprod)>0:
                #this is delta for all users
                userdelta = self.graphc[:,effectprod].dot(prodchange[effectprod])
                yuserdelta = userdelta[Ylist]
                for u in yuserdelta.nonzero()[0]:
                    uidx = Ylist[u]
                    MT.changeVal(uidx,yuserdelta[u])
            'delete next user, make current to next'
            self.A = self.Y
            sizeA -= 1
            setA = setY
            self.fbs = self.yfbs
            self.bsusps = self.ybsusps
            self.vx = vy
            self.vxs.append(self.vx)
            if i % (sizeA0//100 + 1) == 0:
                sys.stdout.write('.')
                sys.stdout.flush()
            i+=1
        print()
        return np.sum(self.A)

    def initfastgreedy(self, ptype, numSing, rbd='avg', eps=1.6):
        '''
        default: ptype=[Ptype.freq], numSing=10, rbd='avg'
        '''
        self.ptype=ptype
        self.numSing=numSing #number of singular vectors we consider
        self.avgexponents=[]
        if len(ptype)==1:
            self.initfastgreedy2D(numSing, rbd, eps=eps)
        elif len(ptype) > 1:
            self.initfastgreedyMD(numSing, rbd, eps=eps)

        self.bestvx = -1
        self.qchop=False
        #reciprocal of indegrees
        self.sindegreciprocal = csr_matrix(self.windegrees).astype(np.float64)
        data = self.sindegreciprocal.data
        nozidx = data.nonzero()[0]
        self.sindegreciprocal.data[nozidx] = data[nozidx]**(-1)

        return

    def tenormatricization(self, tspim, ratepim, tbindic, rbins,
                           mtype=coo_matrix, dropweight=True, logdegree=False):
        'matricize the pim of ts and rates into matrix'
        if tspim is None and ratepim is None:
            return self.graph, range(self.nV)
        tscm, rtcm, dl = None, None,0
        if Ptype.ts in self.ptype and tspim is not None:
            tscm = tspim.edgeidxm.tocoo()
            dl = len(tscm.data)
        if Ptype.rate in self.ptype and ratepim is not None:
            rtcm = ratepim.edgeidxm.tocoo()
            dl = len(rtcm.data)
        if dropweight is True and tspim is not None:
            w = np.multiply(tspim.dropfalls, tspim.dropslops)
            w = np.log1p(w) + 1
        else:
            w = np.ones(self.nV)
        xs, ys, data, colWeights = [],[],[],[] # for matricized tenor
        matcols, rindexcols={},{}
        for i in range(dl):
            if tscm is not None and rtcm is not None:
                assert(tscm.row[i] == rtcm.row[i] and tscm.col[i] == rtcm.col[i])
                u = tscm.row[i]
                v = tscm.col[i]
                for t1, r1 in zip(tspim.eprop[tscm.data[i]],
                                ratepim.eprop[rtcm.data[i]]):
                    t = t1//int(tbindic[self.tunit])
                    r = rbins(r1)
                    strcol = ' '.join(map(str,[v,t,r]))
                    if strcol not in matcols:
                        idx = len(matcols)
                        matcols[strcol] = idx
                        rindexcols[idx]=strcol
                    xs.append(u)
                    ys.append(matcols[strcol])
                    data.append(1.0)
            elif tscm is not None:
                u = tscm.row[i]
                v = tscm.col[i]
                for t1 in tspim.eprop[tscm.data[i]]:
                    t = t1//int(tbindic[self.tunit])
                    strcol = ' '.join(map(str,[v,t]))
                    if strcol not in matcols:
                        idx = len(matcols)
                        matcols[strcol] = idx
                        rindexcols[idx]=strcol
                    xs.append(u)
                    ys.append(matcols[strcol])
                    data.append(1.0)
            elif rtcm is not None:
                u = rtcm.row[i]
                v = rtcm.col[i]
                for r1 in ratepim.eprop[rtcm.data[i]]:
                    r = rbins(r1)
                    strcol = ' '.join(map(str,[v,r]))
                    if strcol not in matcols:
                        idx = len(matcols)
                        matcols[strcol] = idx
                        rindexcols[idx]=strcol
                    xs.append(u)
                    ys.append(matcols[strcol])
                    data.append(1.0)
            else:
                print('Warning: no ts and rate for matricization')
                return self.graph, range(self.nV)

        nrow, ncol = max(xs)+1, max(ys)+1
        sm = mtype( (data, (xs, ys)), shape=(nrow, ncol), dtype=np.float64 )
        if logdegree:
            print('using log degree')
            sm.data[0:] = np.log1p(sm.data)
        if dropweight:
            m1, n1 = sm.shape
            for i in range(n1):
                pos = rindexcols[i].find(' ')
                v = int(rindexcols[i][:pos])
                colWeights.append(w[v])
            colDiag = lil_matrix((n1, n1))
            colDiag.setdiag(colWeights)
            sm = sm * colDiag.tocsr()
        return sm, rindexcols

    def initfastgreedyMD(self, numSing, rbd, eps = 1.6):
        '''
            use matricizationSVD instead of freq matrix svd
        '''
        #afile = self.tsprop if self.tsprop is not None else self.rateprop
        #ipath =  os.path.dirname(os.path.abspath(afile))
        tbindic={}
        if isinstance(self.tsprop, str) and os.path.isfile(self.tsprop):
            tbindic={'s':24*3600, 'd':30}
            print('Generate tensorfile with tunit:{}, tbins:{}'.format(self.tunit,
                                                                   tbindic[self.tunit]))
        elif isinstance(self.tsprop, STensor):
            tbindic={'s':1, 'd':1}
            print('Generate tensorfile with time rescale: ', tbindic[self.tunit] )

        'edgepropertyAnalysis has already digitized the ratings'
        rbins = lambda x: int(x) #lambda x: 0 if x<2.5 else 1 if x<=3.5 else 2
        if self.matricizetenor is None:
            matricize_start = time.clock()
            sm, rindexcol = self.tenormatricization(self.tspim, self.ratepim,
                    tbindic, rbins, mtype=coo_matrix,
                    dropweight=self.priordropslop,
                    logdegree=False)
            self.matricizetenor = sm
            print('::::matricize time cost: ', time.clock() - matricize_start)
        sm = self.matricizetenor
        print("matricize {}x{} and svd dense... ..."\
                .format(sm.shape[0], sm.shape[1]))
        u, s, vt = slin.svds(sm, k=numSing, which='LM')
        u = np.fliplr(u)
        s = s[::-1]
        CU, CV = [],[]
        for i in range(self.numSing):
            ui = u[:, i]
            si = s[i]
            if abs(max(ui)) < abs(min(ui)):
                ui = -1*ui
            if type(rbd) is float:
                sqrtSi = math.sqrt(si)
                ui *= sqrtSi
                rbdrow= rbd
            elif rbd == 'avg':
                rbdrow = 1.0/math.sqrt(self.nU)
            else:
                print('unkown rbd {}'.format(rbd))
            rows = np.argsort(-ui, axis=None, kind='quicksort')
            for jr in range(len(rows)):
                r = rows[jr]
                if ui[r] <= rbdrow:
                    break
            self.avgexponents.append(math.log(jr, self.nU))
            'consider the # limit'
            if self.nU > 1e6:
                e0 = self.e0
                ep = max(eps, 2.0/(3-e0))
                nn = sm.shape[0] + sm.shape[1]
                nlimit = int(math.ceil(nn**(1/ep)))
                cutrows = rows[:min(jr,nlimit)]
            else:
                cutrows = rows[:jr]

            CU.append(cutrows)

        self.CU = np.array(CU)
        self.CV = np.array(CV)
        return

    def initfastgreedy2D(self, numSing, rbd, eps=1.6):
        'rbd threshold that cut the singular vecotors, default is avg'
        'parameters for fastgreedy'
        u, s, vt = slin.svds(self.graphr.astype(np.float64), k=numSing, which='LM')
        #revert to make the largest singular values and vectors in the front
        u = np.fliplr(u)
        vt = np.flipud(vt)
        s = s[::-1]
        self.U = []
        self.V = []
        self.CU = []
        self.CV = []
        for i in range(self.numSing):
            ui = u[:, i]
            vi = vt[i, :]
            si = s[i]
            if abs(max(ui)) < abs(min(ui)):
                ui = -1*ui
            if abs(max(vi)) < abs(min(vi)):
                vi = -1*vi
            if type(rbd) is float:
                sqrtSi = math.sqrt(si)
                ui *= sqrtSi
                vi *= sqrtSi
                rbdrow, rbdcol = rbd, rbd
            elif rbd == 'avg':
                rbdrow = 1.0/math.sqrt(self.nU)
                rbdcol = 1.0/math.sqrt(self.nV)
            else:
                print('unkown rbd {}'.format(rbd))
            rows = np.argsort(-ui, axis=None, kind='quicksort')
            cols = np.argsort(-vi, axis=None, kind='quicksort')
            for jr in range(len(rows)):
                r = rows[jr]
                if ui[r] <= rbdrow:
                    break
            self.avgexponents.append(math.log(jr, self.nU))
            if self.nU > 5e5:
                e0=self.e0
                ep = max(eps, 2.0/(3-e0))
                nn = self.nU + self.nV
                nlimit = int(math.ceil(nn**(1.0/ep)))
                cutrows = rows[:min(jr,nlimit)]
            else:
                cutrows = rows[:jr]
            for jc in range(len(cols)):
                c = cols[jc]
                if vi[c] <= rbdcol:
                    break
            cutcols = cols[:jc]
            'begin debug'
            self.U.append(ui)
            self.V.append(vi)
            'end debug'
            self.CU.append(cutrows)
            self.CV.append(cutrows)

        self.CU = np.array(self.CU)
        self.CV = np.array(self.CV)
        return

    def qfunc(self, ratios, fbs=None, scale='exp', numratios=1):
        if self.aggmethod == 'rank':
            'do not use qfun if it is rank aggregation'
            return ratios

        if self.suspbd <= 0.0:
            greatbdidx = ratios > 0.0
        else:
            greatbdidx = ratios >= self.suspbd
            lessbdidx = ratios < self.suspbd
            'picewise q funciton if < suspbd, i.e. epsilon1'
            ratios[lessbdidx] = 0.0
        'picewise q funciton if >= suspbd, i.e. epsilon1'
        if scale == 'exp':
            ratios[greatbdidx] = self.expbase**(ratios[greatbdidx]-numratios)
        elif scale == 'pl':
            ratios[greatbdidx] = ratios[greatbdidx]**self.b
        elif scale == 'lin':
            ratios[greatbdidx] = np.fmax(self.b*(ratios[greatbdidx]-1)+1, 0)
        else:
            print('unrecognized scale: ' + scale)
            sys.exit(1)
        return ratios

    def setdeltapimsusp(self, z, ysuspusers, add):
        if Ptype.ts in self.ptype:
            self.tspim.deltasuspects(z, ysuspusers, add)
        if Ptype.rate in self.ptype:
            self.ratepim.deltasuspects(z, ysuspusers, add)
        return

    def removecurrentblock(self, rows):
        '''it is for find second block, remove rows from
           self.graph, self.matricizetenor
        '''
        print('removing {} rows from graph'.format(len(rows)))
        lilm = self.graph.tolil()
        lilm[rows,:]=0
        self.graph=lilm.tocoo()
        self.graphc= lilm.tocsc()
        self.graphr = self.graph.tocsr()

        if self.matricizetenor is not None:
            print('removing {} rows from tensor'.format(len(rows)))
            lilmm = self.matricizetenor.tolil()
            lilmm[rows,:] = 0
            self.matricizetenor = lilmm.tocoo()
        return

    #@profile
    def fastgreedy(self):
        'adding and deleting greed algorithm'
        'No Need: user order for r with obj fuct'
        self.fastlocalbest = []
        self.fastbestvx = 0
        self.fastbestA, self.fastbestfbs, self.fastbestbsusps = \
                np.zeros(self.nU), np.zeros(self.nV), np.zeros(self.nV)
        for k in range(self.numSing):
            print('process {}-th singular vector'.format(k+1))
            lenCU = len(self.CU[k])
            if lenCU == 0:
                continue
            print('*** *** shaving ...')
            A0 = np.zeros(self.nU, dtype=int)
            A0[self.CU[k]]=1 #shaving from sub singluar space
            self.start(A0, ptype=self.ptype)
            self.greedyshaving()
            print('*** *** shaving opt size: {}'.format(sum(self.bestA)))
            print('*** *** shaving opt value: {}'.format(self.bestvx))
            if self.fastbestvx < self.bestvx:
                self.fastbestvx = self.bestvx
                self.fastbestA = np.array(self.bestA)
                self.fastbestfbs = np.array(self.bestfbs)
                self.fastbestbsusps = np.array(self.bestbsusps)
                print('=== === improved opt size: {}'.format(sum(self.fastbestA)))
                print('=== === improved opt value: {}'.format(self.fastbestvx))

            brankscores = np.multiply(self.bestbsusps, self.bestfbs)
            A = self.bestA.nonzero()[0]
            self.fastlocalbest.append((self.bestvx, (A, brankscores)))
            'clear shaving best'
            self.bestvx = 0

        self.bestvx, self.bestA, self.bestfbs, self.bestbsusps = \
                    self.fastbestvx, self.fastbestA, \
                    self.fastbestfbs, self.fastbestbsusps
        return

    def drawObjectiveCurve(self, outfig):
        import matplotlib.pyplot as plt
        fig = plt.figure()
        plt.plot(self.vxs, '-')
        plt.title('The convergence curve of simulated anealing.')
        plt.xlabel('# of iterations')
        plt.ylabel('objective value')
        if outfig is not None:
            fig.savefig(outfig)
        return fig


def holoscope_interface(wmat, alg, ptype, qfun, b, rateprop=None, tsprop=None,
              tunit='s', numSing=10, nblock=1, eps=1.6):
    '''
    The interface of HoloScope algorithm for external use
    Parameters
    ----------
    wmat: str or sparse matrix
        If it is str, wmat is the input file name. We load the file into sparse
        matrix. If it is sparse matrix, we just use wmat.
    alg: str
        which algorithm you are going to use. You can choose 'greedy' for
        synthetic data (#rows+#cols<10000); or 'fastgreedy' for any size of data
        sets.
    ptype: list
        contains which attributes the algorithm is going to use. The hololisc
        use of all siginals is [Ptype.freq, Ptype.ts, Ptype.rate]
    qfun: str
        which kind of qfun the algorithm uses, choosing from 'exp' for
        exponential (recommended), 'pl' for power-law, 'lin' for linear
    b: float
        The base of exponetial qfun, or the exponent of power-law qfun, or
        absolute slope of linear qfun
    rateprop: str or STensor or None
        The file name with path for user-object rating sequences. The file
        format is that each line looks like 'userid-objectid:#star1 #star2
        ...\n'.
        If it is STensor, then rateprop contains (userid, objectid, #star) --> freq
    tsprop: str or None
        The file name with path for user-object timestamp sequences. The file
        format is that each line looks like 'userid-objectid:t1 t2 ...\n'
        If it is STensor, then rateprop contains (userid, objectid, tsbin) --> freq
    tunit: str (only support 's' or 'd') or None
        The time unit of input time
        e.g. in amazon and yelp data, the time is date, i.e. tunit='d'.
             We use # of days (integer) from the earlest date as input
        It does not need if tsprop is STensor
    numSing: int
        The number of first left singular vectors used in our algorithm
    nblock: int
        The number of block we need from the algorithm
    eps: float
        It gives the approximate level cut for singular vectors, which
        is a trade-off parameter for efficency and accuracy. Usually eps
        is between (1.5, 2], and the complexity reduce from the quadratic in number of
        nodes to the near linear in number of edges.
    Return
    ---------
    (gbestvx, (gsrows, gbscores)), opt
        Block (gsrows, gbscores) has the best objective values 'gbestvx' among
	*nblock* blocks.
	gbestvx: float
            the best objective value of the *nblock* blocks.
        gsrows: list
            is the list of suspicious rows.
        gbscores: list
            is the suspicoius scores for every objects. The index is object id,
            and value is the score. With the scores, you can get the suspicious rank
        of the objects.
        opt: instance of HoloScopeOpt class
            the class instance which contains all the *nblock* blocks in opt.nbests.
            opt.nbests: list
                This is the list contains *nblock* solutions in the form of
                tuple, i.e., (opt.bestvx, (srows, bscores))
    '''
    print('initial...')
    if sci.sparse.issparse(wmat) is False and os.path.isfile(wmat):
        # sm = loadedge2sm(wmat, coo_matrix, weighted=True, idstartzero=True)
        # remove unexpected parameter 'weighted'
        sm = loadedge2sm(wmat, coo_matrix, idstartzero=True)
    else:
        sm = wmat.tocoo()
    inprop = 'Considering '
    if Ptype.freq in ptype:
        inprop += '+[topology] '
    if Ptype.ts in ptype:
        inprop += '+[timestamps] '
        #consider sdrop by default when Ptype.ts
        inprop += '+[sudden drop]'
    else:
        tsprop=None
    if Ptype.rate in ptype:
        inprop += '+[rating i.e. # of stars] '
    else:
        rateprop = None
    print(inprop)

    opt = HoloScopeOpt(sm, qfun=qfun, b=b, tsprop=tsprop, tunit=tunit,
            rateprop=rateprop)
    opt.nbests=[]
    opt.nlocalbests=[] #mainly used for fastgreedy
    gsrows,gbscores,gbestvx = 0,0,0
    for k in range(nblock):
        start_time = time.clock()
        if alg == 'greedy':
            n1, n2 = sm.shape
            if n1 + n2 > 1e4:
                print('[Warning] alg {} is slow for size {}x{}'\
                        .format(alg, n1, n2))
            A = np.ones(opt.nU,dtype=int)
            print('initial start')
            opt.start(A, ptype=ptype)
            print('greedy shaving algorithm ...')
            opt.greedyshaving()
        elif alg == 'fastgreedy':
            print("""alg: {}\n\t+ # of singlular vectors: {}\n""".format(alg, numSing))
            print('initial start')
            opt.initfastgreedy( ptype, numSing, eps=eps )
            print("::::Finish Init @ ", time.clock() - start_time)
            print('fast greedy algorithm ...')
            opt.fastgreedy()
            opt.nlocalbests.append(opt.fastlocalbest)
        else:
            print('No such algorithm: '+alg)
            sys.exit(1)

        print("::::Finish Algorithm @ ", time.clock() - start_time)

        srows = opt.bestA.nonzero()[0]
        bscores = np.multiply(opt.bestfbs, opt.bestbsusps)
        opt.nbests.append((opt.bestvx, (srows, bscores)))
        gsrows, gbscores, gbestvx = (srows,bscores,opt.bestvx) \
                if gbestvx < opt.bestvx  else (gsrows, gbscores, gbestvx)
        if k < nblock-1:
            opt.removecurrentblock(srows)

    #levelcols = score_level_objects( gbscores )
    nnzcols = nonzero_objects( gbscores )

    #print('global best size: nodes', len(gsrows), len(nnzcols), 'with camouflage.')

    #print('global best value ', gbestvx)

    return ((gsrows, nnzcols), gbestvx), gbscores[nnzcols], opt


class HoloScope( DMmodel ):
    '''Anomaly detection base on contrastively dense subgraphs, considering
    topological, temporal, and categorical (e.g. rating scores) signals, or
    any supported combinations.

    Parameters
    ----------
    graph: Graph
        Graph instance contains adjency matrix, and possible multiple signals.
    alg: string options ['fastgreedy' | 'greedy' ]
        The algorithm used for detect dense blocks. You can choose 'greedy' for
        synthetic data (#rows+#cols<10000); or 'fastgreedy' for any size of data
        sets.
        Default is 'fastgreedy' which
        uses main (with first several large singular values) truncated singular vectors
        to find dense blocks. alg is used with eps as truncation factor, and
        numSing as number of first large singular vectors.
    eps: float
        It gives the approximate level cut for singular vectors, which
        is a trade-off parameter for efficency and accuracy. Usually eps
        is between (1.5, 2], and the complexity reduce from the quadratic in number of
        nodes to the near linear in number of edges.
        Larger eps gives faster detection, but may miss the denser blocks.
        Default is 1.6.
    numSing: int
        The number of first large left singular vectors used in our algorithm
    qfun: string options ['exp' | 'pl' | 'lin']
        which kind of qfun the algorithm uses, choosing from 'exp' for
        exponential (recommended), 'pl' for power-law, 'lin' for linear
        Default is 'exp'.
    b: float
        The base of exponetial qfun, or the exponent of power-law qfun, or
        absolute slope of linear qfun
        Default is 32.
    '''
    def __init__(self, graph, **params):
        self.graph = graph
        self.alg = param_default(params, 'alg', 'fastgreedy')
        self.eps = param_default(params, 'eps', 1.6)
        self.numSing = param_default(params, 'numSing', 10)
        self.qfun = param_default(params, 'qfun', 'exp')
        self.b = param_default(params,'b', 32)

    def __str__(self):
        return str(vars(self))


    def run(self, k:int=1, level:int=0, eps:float = 1.6):
        '''run with how many blocks are output.
        Parameters:
        --------
        nblock: int
            The number of block we need from the algorithm
        level: int
            The level of signals used for anomaly detection. Choose in [0 | 1 | 2 |
            3]. 0: topology only. 1: topology with time. 2: topology with category
            (e.g. rating score). 3: all three.
            Default is 0.
        '''

        if eps != 1.6:
            epsuse=1.6
        else:
            epsuse = self.eps
        self.level = level
        nprop = self.graph.nprop
        graph = self.graph

        tsprop, rateprop = None, None
        if self.level == 0:
            ptype=[Ptype.freq]
        elif self.level ==1:
            ptype=[Ptype.freq, Ptype.ts]
            if nprop < 1:
                raise Exception("Error: at least 3-mode graph tensor is needed for level 1")
            tsprop = graph.get_time_tensor()
        elif self.level == 2:
            ptype = [Ptype.freq, Ptype.rate]
            if nprop < 1:
                raise Exception("Error: at least 3-mode graph tensor is needed for level 2")
            "The order of mode in graph tensor for categorical bins if exit, start from zero."
            modec = 3 if nprop > 1 else 2
            rateprop = graph.get_one_prop_tensor(modec)
        elif self.level == 3:
            ptype = [Ptype.freq, Ptype.ts, Ptype.rate]
            if nprop < 2:
                raise Exception("Error: at least 4-mode graph tensor is needed for level 3")
            tsprop = graph.get_time_tensor()
            modec=3
            rateprop = graph.get_one_prop_tensor(3)
        else:
            print("Warning: no run level ",self.level,", use level 0 instead!")
            ptype=[Ptype.freq]

        bdres = holoscope_interface(graph.sm.astype(float),
                self.alg, ptype, self.qfun, self.b,
                tsprop=tsprop, rateprop=rateprop,
                nblock=k, eps=epsuse, numSing=self.numSing)

        nres = []
        opt = bdres[-1]
        for nb in range(k):
            res = opt.nbests[nb]
            print('block{}: \n\tobjective value {}'.format(nb + 1, res[0]))
            levelcols = score_level_objects(res[1][1])
            nnzcols = nonzero_objects(res[1][1])
            rows = res[1][0]
            nleveledges = graph.get_subgraph_nedges( rows, levelcols )
            nedges = graph.get_subgraph_nedges( rows, nnzcols )
            print( '\tNode size: {} x {}, edge size {}'.format(len(rows),
                len(levelcols), nleveledges) )
            print( '\tRow and nonzero columns {} x {}, edge size: {} with camouflage.'.format(
                len(rows), len(nnzcols), nedges) )
            nres.append( ( (rows, levelcols), res[0], nnzcols, res[1][1][nnzcols] ) )

        self.nres = nres

        return nres

    def anomaly_detection(self, k:int=1, eps:float = 1.6):
        return self.run(k=k, eps=eps)

    def save(self, outpath):
        import pickle
        out = open(outpath,'wb')
        pickle.dump(self.nres, out)
        pass





#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

############################################################################
# Beyond outliers and on to micro-clusters: Vision-guided Anomaly Detection
# Authors: Wenjie Feng, Shenghua Liu, Christos Faloutsos, Bryan Hooi,
#                 and Huawei Shen, and Xueqi Cheng
#
#  Project: eaglemine
#      eaglemine_model.py
#      Version:  1.0
#      Date: Dec. 17 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <12/17/2017>
#
#      Main contributor:  Wenjie Feng.
###########################################################################

__author__ = 'wenchieh'

# sys
import os
import operator
from collections import deque

# third-party lib
import numpy as np

# project
from .core.leveltree import LevelTree
from .utils.loader import Loader
from .utils.mdlbase import MDLBase
from .desc.dtmnorm_describe import DTMNormDescribe
from .desc.truncated_gaussian import TruncatedGaussian
from .desc.gaussian_describe import GaussianDescribe
from .desc.normal_gaussian import NormalGaussian
from .desc.statistic_hypothesis_test import StatisticHypothesisTest
from .desc.statistic_hypothesis_test_truncate import StatisticHypothesisTestTruncate


tiny_blobs = 'tiny_blob2cnt.out'
contracttree = 'level_tree_contract.out'
prunetree = 'level_tree_prune.out'
refinetree = 'level_tree_refine.out'


class EagleMineModel(object):
    _valid_vocabularies_ = ['dtmnorm', 'dmgauss']

    def __init__(self, mode=2, mix_comps=2, power_base=2):
        self.leveltree = None
        self.describes = None
        self.mode = mode
        self._hcel = None
        self._count = None
        self._total = None
        self.shape = None
        self.descvoc = None
        self._mixcomps = mix_comps
        self.power_base = power_base

    def set_vocabulary(self, desc_voc='dtmnorm'):
        # assert desc_voc in self._valid_vocabularies_
        if desc_voc == 'dtmnorm':
            self.descvoc = {"name": 'dtmnorm', "voc": DTMNormDescribe, "dist": TruncatedGaussian}
        elif desc_voc == 'dmgauss':
            self.descvoc = {"name": 'dmgauss', "voc": GaussianDescribe, "dist": NormalGaussian}
        else:
            raise ValueError("Unimplemented initialization method {}, "
                             "valid vocabularies {}".format(desc_voc, self._valid_vocabularies_))

    # def load_histogram(self, infn_hist):
    #     loader = Loader()
    #     shape, ticks, hist_arr = loader.load_histogram(infn_hist)
    #     self._hcel = hist_arr[:, :-1]
    #     self._count = hist_arr[:, -1]
    #     n_cel, n_dim = self._hcel.shape
    #     if n_dim != self.mode:
    #         raise ValueError("Input histogram dimension does match with the initial dimension.")
    #     self._total = n_cel
    #     self.shape = shape

    def set_histogram(self, hcel2cnt):
        self._hcel = np.asarray(hcel2cnt[:, :-1])
        self._count = np.asarray(hcel2cnt[:, -1])
        n_pos, n_dim = self._hcel.shape
        if n_dim != self.mode:
            raise ValueError("Input histogram dimension does match with the initial dimension.")
        self._total = n_pos
        self.shape = np.max(self._hcel, axis=0)

    def load_leveltree(self, infn, sep=',', verbose=True):
        self.leveltree = LevelTree()
        self.leveltree.load_leveltree(infn, sep, verbose)

    def load_describes(self, infn, sep=';', P=2, verbose=True):
        self.describes = dict()
        hcel2cnt = dict()
        for k in range(self._total):
            hcel2cnt[tuple(self._hcel[k])] = self._count[k]

        with open(infn, 'r') as ifp:
            for line in ifp:
                line = line.strip()
                splits = line.find(sep)
                node_id = int(line[:splits])
                desc = self.descvoc["voc"](self.mode)
                desc.load(line[splits+1:])
                tree_node = self.leveltree.get_node(node_id)
                node_counts = list()
                for hc in tree_node.get_covers():
                    node_counts.append(hcel2cnt.get(tuple(hc)))
                values = np.array(node_counts)
                values = np.log(values + 1) / np.log(P)
                desc.set_data(tree_node.get_covers(), values)
                self.describes[node_id] = desc
            ifp.close()

        if verbose:
            print('load describe done!')

    def dump(self):
        print("Information: ")
        print("#mode: {}, input-shape: {}".format(self.mode, self.shape))
        print("#non-zeros: {}, #totals: {}".format(len(self._hcel), np.sum(self._count)))
        print("describe vocabulary: {}".format(self.descvoc))
        print("Level-tree information:")
        self.leveltree.dump()
        print("Leaves describe information:")
        for id, desc in self.describes.items():
            print("{}: {}".format(id, str(desc)))

        #print("done!\n")

    def save(self, outfn, delimiter=";"): # , suffix=''
        # describe_outfn = 'describe%s.out' % suffix
        # os.path.join(outfd, describe_outfn)
        # self.leveltree.save_leveltree(outfd+'level_tree_expand-00.out')
        # print("save done!\n")
        with open(outfn, 'w') as ofp:
            desc_ids = sorted(self.describes.keys())
            for id in desc_ids:
                desc = self.describes[id]
                line = str(id) + delimiter + str(desc)
                ofp.writelines(line + '\n')
            ofp.close()

    def leveltree_build(self, outfd, step=0.2, prune_alpha=0.95, min_level=1.0, max_level=None, verbose=True):
        values = np.log2(1.0 + np.asarray(self._count, float))
        if max_level is None or max_level < min_level or max_level > np.max(values):
            max_level = np.max(values)

        print("\n========== ")
        print("Construct raw-tree.")
        self.leveltree = LevelTree()
        self.leveltree.build_level_tree(self._hcel, values, min_level, max_level, step,
                                        verbose=verbose, outfn = os.path.join(outfd, tiny_blobs))
        print("\n========== ")
        print("Refine tree structure.")
        print("\n+++++++++ ")
        print("a). tree contract")
        self.leveltree.tree_contract(verbose=verbose)
        if verbose:
            print("Info: Contract level-tree:")
            self.leveltree.dump()
        print("\n+++++++++ ")
        print("b). tree prune")
        self.leveltree.tree_prune(alpha=prune_alpha, verbose=verbose)
        if verbose:
            print("Info: Pruned level-tree:")
            self.leveltree.dump()
        print("\n+++++++++ ")
        print("c). tree node expand")
        self.leveltree.tree_node_expand(verbose)
        # self.leveltree.save_leveltree(os.path.join(outfd, refinetree))
        if verbose:
            print("Info: Expanded level-tree:")
            self.leveltree.dump()
        print("\ndone")

    def _describe_singular_check_(self, hcels):
        hcels = np.asarray(hcels)
        ndims = hcels.ndim
        sing_dims = list()
        for dm in range(ndims):
            if len(np.unique(hcels[:, dm])) <= 1:
                sing_dims.append(dm)

        if len(sing_dims) > 0:
            raise ValueError("{}-dimensional data degenerate in "
                             "{}-st dimension(s)".format(ndims, sing_dims))

    def describe_all(self):
        self.describes = dict()
        hcel2cnt = dict()
        for k in range(self._total):
            hcel2cnt[tuple(self._hcel[k])] = self._count[k]

        # tree nodes will be fitted with mixture distribution
        heavynode2level = self.leveltree.get_heavynodes()
        nodes = self.leveltree.get_nodes()  # self.leveltree.get_leaves()
        for ndid in nodes:
            tree_node = self.leveltree.get_node(ndid)
            hcels = tree_node.get_covers()
            node_counts = [hcel2cnt.get(tuple(hc)) for hc in hcels]
            is_mix = ndid in heavynode2level.keys()

            ## logP_count
            values = np.array(node_counts)
            values = np.log(values + 1) / np.log(self.power_base)

            comps = self._mixcomps if is_mix else 1
            desc = self.descvoc["voc"](self.mode, is_mix, comps)
            if self.descvoc["name"] == 'dtmnorm':
                desc.set_bounds()

            self._describe_singular_check_(hcels)
            desc.fit(hcels, values)
            self.describes[ndid] = desc

    def _model_hcels_prob_(self, hcel_left, hcel_right, paras, is_mix):
        npts = len(hcel_left)
        mus, covs, weights = paras["mus"], paras["covs"], paras["weights"]
        probs = None
        if is_mix:
            nmix = len(mus)
            mix_dists = list()
            for i in range(nmix):
                if self.descvoc["name"] == "dtmnorm":
                    desc = list(self.describes.values())[0]
                    lower_bnd, upper_bnd = desc.get_bounds()
                    idist = self.descvoc["dist"](lower_bnd, upper_bnd)
                else:
                    idist = self.descvoc["dist"]()
                idist.set_para(mus[i], covs[i])
                mix_dists.append(idist)

            comp_probs = list()
            for k in range(npts):
                ps = [mix_dists[i].range_cdf(hcel_left[k, :], hcel_right[k, :]) for i in range(nmix)]
                comp_probs.append(ps)
            probs = np.array(comp_probs)
        else:
            if self.descvoc["name"] == "dtmnorm":
                desc = list(self.describes.values())[0]
                lower_bnd, upper_bnd = desc.get_bounds()
                dist = self.descvoc["dist"](lower_bnd, upper_bnd)
            else:
                dist = self.descvoc["dist"]()
            dist.set_para(mus[0], covs[0])
            probs = np.array([dist.range_cdf(hcel_left[k, :], hcel_right[k, :]) for k in range(npts)])

        return probs

    def search(self, min_pts=20, strictness=4, verbose=True):
        search_tree = dict()
        blob_nodes = list()
        hcel2cnt = dict()
        min_pts = np.min([min_pts, int(np.mean(self._count))])
        for k in range(self._total):
            hcel2cnt[tuple(self._hcel[k])] = self._count[k]

        heavynode2level = self.leveltree.get_heavynodes()

        stat_tester = None
        if self.descvoc['name'] == 'dtmnorm':
            stat_tester = StatisticHypothesisTestTruncate(alpha_level=0.01, n_jobs=3)
        elif self.descvoc['name'] == 'dmgauss':
            stat_tester = StatisticHypothesisTest(strictness)
        else:
            raise ValueError("Unimplemented vocabulary, valid vocabularies {}".format(self._valid_vocabularies_))

        Q = deque()
        roots = self.leveltree.get_nodesid_atlevel(self.leveltree.levels[0])
        Q.extend(roots)

        # BFS search
        if self.describes is None:
            self.describes = dict()
        while len(Q) > 0:
            ndid = Q.popleft()
            tree_node = self.leveltree.get_node(ndid)
            hcels = np.array(tree_node.get_covers())
            node_counts = [hcel2cnt.get(tuple(hc)) for hc in hcels]

            if np.max(node_counts) < min_pts:
                blob_nodes.append(tree_node)

            is_mix = ndid in heavynode2level.keys()

            ## logP
            values = np.array(node_counts)
            values = np.log(values + 1) / np.log(self.power_base)

            self._describe_singular_check_(hcels)
            comps = self._mixcomps if is_mix else 1
            if ndid not in self.describes:
                desc = self.descvoc["voc"](self.mode, is_mix, comps)
                if self.descvoc["name"] == 'dtmnorm':
                    desc.set_bounds()
                desc.fit(hcels, values)
                self.describes[ndid] = desc
            else:
                desc = self.describes[ndid]
                desc.data, desc.values = hcels, values

            hcels_prob = None
            if is_mix:
                hcels_prob = self._model_hcels_prob_(hcels, hcels + 1, desc.paras, is_mix)

            weights = values #np.ones(len(hcels), int) #
            gaussian = False
            if self.descvoc['name'] == 'dmgauss':
                gaussian = stat_tester.apply(hcels, weights, hcels_prob, desc.paras, is_mix)
            elif self.descvoc['name'] == 'dtmnorm':
                lower_bnd, _ = desc.get_bounds()
                gaussian = stat_tester.apply(hcels, weights, hcels_prob, desc.paras, is_mix, lower_bnd)

            if gaussian:
                search_tree[ndid] = desc
                if verbose:
                    print("Info: island: {} hypothesis * Accept. ^~^".format(ndid))
            else:
                if tree_node.child is not None:
                    Q.extend(tree_node.child)
                    if verbose:
                        print("Info island: {} hypothesis & Rejected. >_<".format(ndid))
                else:
                    search_tree[ndid] = desc
                    if verbose:
                        print("Info: island: {} hypothesis $ pseudo-Accept. ^..^".format(ndid))

        self.describes = search_tree

    def _close_check(self, gaus_voc1, gaus_voc2, threshold=None):
        mu1, cov1 = np.array(gaus_voc1.get("mus")[0]), np.array(gaus_voc1.get("covs")[0])
        mu2, cov2 = np.array(gaus_voc2.get("mus")[0]), np.array(gaus_voc2.get("covs")[0])

        diff = mu1 - mu2 # distance between two cluster centers.
        distance = np.sqrt(diff.dot(diff.T))
        if threshold is not None:
            return distance <= threshold
        else:
            # default distance threshold
            cov_dist = np.max([1, np.max(np.sqrt(np.diag(cov1)) + np.sqrt(np.diag(cov2)))])
            return distance < 2 * cov_dist  #3 * np.sum(axes_distance)

    def _greedy_select(self, candidates, content):
        '''
        select the optimal merged cluster with max score (minimum decrease of log-likelihood)
        :param candidates: selection candidates
        :param clusters: all clusters dictionary
        :return:  merged cluster id
        '''
        if len(candidates) <= 0: return None

        cands_score = list()
        for cand in candidates:
            cs = content.get(cand)
            if len(cs.get("cs_id")) > 0:
                score = 0.0
                for cs_id in cs.get("cs_id"):
                    desc = content[cs_id].get("desc")
                    score += desc.paras["loss"]
                score -= cs.get("desc").paras["loss"]
                score /= 1.0 * cs.get("npts")            # average score over points
                # log-likelihood will decrease (need to select the one keep best model.)
                cands_score.append((cand, score))
        sorted_score = sorted(cands_score, cmp=lambda x, y: x[1]-y[1], reverse=False)
        return sorted_score[0][0]

    def post_stitch(self, strictness=4, verbose=True):
        optimals = list()
        mixturec, singlesc = None, dict()
        hcel2cnt = dict()
        for k in range(self._total):
            hcel2cnt[tuple(self._hcel[k])] = self._count[k]

        ider = -1
        for cid, desc in self.describes.items():
            if cid > ider: ider = cid
            if desc.is_mix: mixturec = cid
            else:
                npts = np.sum([hcel2cnt.get(tuple(pos)) for pos in desc.data])
                singlesc[cid] = {"id": cid, "desc": desc, "cs_id": [cid], "islnds": [cid], "npts": npts}
                optimals.append(cid)

        # stitching islands iteratively
        tested_cands = dict()

        stat_tester = None
        if self.descvoc['name'] == 'dtmnorm':
            stat_tester = StatisticHypothesisTestTruncate(alpha_level=0.05, n_jobs=3)
        elif self.descvoc['name'] == 'dmgauss':
            stat_tester = StatisticHypothesisTest(strictness)
        else:
            raise ValueError("Unimplemented vocabulary, valid vocabularies {}".format(self._valid_vocabularies_))

        while True:
            update = False
            candidates = list()
            n_cands = len(optimals)

            # test each clusters pair.
            for i in range(n_cands):
                ci = singlesc.get(optimals[i])
                for j in range(i + 1, n_cands):
                    cj = singlesc.get(optimals[j])
                    # have tested tuple
                    merge_tuple = (ci.get("id"), cj.get("id"))
                    merge_id = tested_cands.get(merge_tuple, None)
                    if verbose:
                        print(merge_tuple)
                    if merge_id is not None:
                        if merge_id != -1:
                            candidates.append(merge_id)
                        else:
                            continue
                    else:
                        desci, descj = ci.get("desc"), cj.get("desc")
                        tested_cands[merge_tuple] = -1
                        # merge limits: two islands are closed.
                        # distance = 0.3 * np.min(self.shape)
                        closed = self._close_check(desci.paras, descj.paras)
                        if not closed: continue

                        ## output testing cases
                        if verbose:
                            print("Info: test new merging node: {} ({}, {})".format(
                                merge_tuple, ci.get("cs_id"), cj.get("cs_id")))

                        merged_hcels = np.vstack([desci.data, descj.data])
                        merged_values = np.hstack([desci.values, descj.values])
                        desc = self.descvoc["voc"](self.mode, False, 1)
                        if self.descvoc["name"] == 'dtmnorm':
                            desc.set_bounds()
                        desc.fit(merged_hcels, merged_values)

                        weights = np.ones(len(merged_hcels), int) # merged_values #

                        gaussian = False
                        if self.descvoc['name'] == 'dmgauss':
                            gaussian = stat_tester.apply(merged_hcels, weights, None, desc.paras, False)
                        elif self.descvoc['name'] == 'dtmnorm':
                            lower_bnd, _ = desc.get_bounds()
                            gaussian = stat_tester.apply(merged_hcels, weights, None, desc.paras, False, lower_bnd)

                        if gaussian:
                            ider += 1
                            cs_id = [ci.get("id")] + [cj.get("id")]
                            islnds = list(set(ci.get("islnds") + cj.get("islnds")))
                            npts = ci.get("npts") + cj.get("npts")
                            singlesc[ider] = {"id": ider, "desc": desc, "cs_id": cs_id, "islnds": islnds, "npts": npts}
                            candidates.append(ider)
                            tested_cands[merge_tuple] = ider
                            update = True
                            if verbose:
                                print("new node: {}-{}-{}".format(ider, cs_id, islnds))
                        else:
                            if verbose:
                                print("Info:  ------ node {} check failed ------".format(merge_tuple))

            # select the most promising merge-node with greedy select from candidates
            #     (minimum decrease of log-likelihood)
            opt_merge_id = self._greedy_select(candidates, singlesc)
            if opt_merge_id is not None:
                update = True
                opt_c = singlesc.get(opt_merge_id)
                for cid in opt_c.get("cs_id"):
                    if cid in optimals:
                        optimals.remove(cid)
                    if cid in singlesc:
                        for islnd_id in singlesc.get(cid).get("islnds"):
                            if islnd_id in optimals:
                                optimals.remove(islnd_id)
                optimals.append(opt_merge_id)
                if verbose:
                    print("current optimals: {}".format(optimals))
            else:
                update = False

            if not update:
                break    # # no further update and stitch finished

        # if verbose:
        #     print("\nFinal optimal cluster result: {}".format(optimals))

        desc_stitch = dict()
        desc_stitch[mixturec] = self.describes.get(mixturec)
        for cid in optimals:
            c_nd = singlesc.get(cid)
            desc_stitch[cid] = c_nd.get("desc")
        self.describes = desc_stitch

    def _measure_model_mdl_(self, hcels, counts, hcels_label, hcels_prob, outlier_marker=-1):
        counts = np.asarray(counts)
        hcels_label = np.asarray(hcels_label)

        descs = list(self.describes.values())
        N_cls = len(descs)

        mdl = MDLBase()
        L_clusters = mdl.integer_mdl(N_cls)
        L_paras, L_assign, L_nis, L_error = 0.0, 0.0, 0.0, 0.0
        for k in range(N_cls):
            lb = k
            index = np.where(hcels_label == lb)
            k_hcels, k_cnts = hcels[index], counts[index]
            k_probs = hcels_prob[index]

            # model error code-length
            knis = np.sum(descs[k].values)
            exp_logPvs = k_probs * knis
            exp_counts = np.array(np.power(self.power_base, exp_logPvs), int) #exp_logPvs  #
            L_error += mdl.seq_diff_mdl(exp_counts, k_cnts)

            # model parameters code-length
            comp_parm = descs[k].compact_parm()
            L_paras += np.sum([mdl.float_mdl(p) for p in comp_parm])
            L_assign += 1.0  # for encoding the indicator for mixture or single
            L_nis += mdl.float_mdl(knis)  # for encoding #pts of this cluster

        L_outs = 0.0
        outs_index = np.where(hcels_label == outlier_marker)
        outs_hcels, outs_vals = hcels[outs_index], counts[outs_index]
        L_outs += mdl.seq_diff_mdl(np.zeros_like(outs_vals, int), outs_vals)
        L_shape = np.sum([mdl.integer_mdl(dm) for dm in self.shape])

        C = L_clusters + L_assign + L_shape + L_paras + L_nis + L_outs + L_error
        return C

    def _smooth_entropy_(self, ps, qs, weights, smoother=1e-15):
        n = len(ps)
        ps = np.asarray(ps) + smoother
        qs = np.asarray(qs) + smoother
        w_en = [ weights[i] * ps[i] * np.log2(ps[i] * 1.0 / qs[i])
                 if qs[i] != 0 else 0 for i in range(n) ]
        return np.sum(w_en)

    def _measure_suspicious_(self, hcels, values):
        majority_normal = None
        cls_id2prob = dict()
        for id, desc in self.describes.items():
            hcel_prob = self._model_hcels_prob_(hcels, hcels + 1, desc.paras, desc.is_mix)
            if desc.is_mix:
                majority_normal = id
                hcel_prob = np.sum(hcel_prob * desc.paras['weights'], axis=1)
            cls_id2prob[id] = hcel_prob

        dists = dict()
        for id, probs in cls_id2prob.items():
            dists[id] = self._smooth_entropy_(probs, cls_id2prob[majority_normal], values)

        return sorted(dists.items(), key=operator.itemgetter(1), reverse=True)

    def _hcels_labeling_(self, hcels, outliers_marker=-1, strictness=4):
        criticals = [1e-1, 5e-2, 1e-2, 1e-3, 1e-4, 1e-5]
        hcels = np.asarray(hcels)
        hcel_clsprob = list()
        hcel_labs = list()

        descs = list(self.describes.values())
        mixId = -1
        for k in range(len(descs)):
            desc = descs[k]
            clsprob = self._model_hcels_prob_(hcels, hcels + 1, desc.paras, desc.is_mix)
            if desc.is_mix:
                mixId = k
                clsprob = np.sum(clsprob * desc.paras['weights'], axis = 1)
            hcel_clsprob.append(clsprob)
        hcel_clsprob = np.array(hcel_clsprob).T

        for k in range(self._total):
            prob, cls = np.max(hcel_clsprob[k, :]), np.argmax(hcel_clsprob[k, :])
            if prob <= criticals[strictness]:
                label = outliers_marker
            else:
                label = cls

            hcel_labs.append(label)

        probs = np.max(hcel_clsprob, axis=1)
        return hcel_labs, probs
    
    def cluster_remarks(self, strictness=4, verbose=True):
        outliers_marker = -1
        hcels_labs, hcels_prob = self._hcels_labeling_(self._hcel, outliers_marker, strictness)
        mdls = self._measure_model_mdl_(self._hcel, self._count, hcels_labs, hcels_prob, outliers_marker)
        hcel2label = dict(zip(map(tuple, self._hcel), hcels_labs))
        # suspicious = self._measure_suspicious_(self._hcel, np.log2(1 + self._count))
        if verbose:
            print("** mdl (Model Description Length) for describing the graph feature: {} **".format(mdls))
            # print("Island suspicious orders: {}".format(suspicious))
        
        return hcel2label, mdls, #suspicious
    
    def cluster_histogram(self, outfn=None, strictness=4, comments="#", delimiter=",", verbose=True):
        outliers_marker = -1
        hcels_label, hcels_prob = self._hcels_labeling_(self._hcel, outliers_marker, strictness)
        mdls = self._measure_model_mdl_(self._hcel, self._count, hcels_label, hcels_prob, outliers_marker)
        # suspicious = self._measure_suspicious_(self._hpos, np.log2(1 + self._count))

        if outfn is not None:
            np.savetxt(outfn, np.vstack((self._hcel.T, hcels_label)).T, '%i', delimiter=delimiter, comments=comments,
                       header='#hcel: {}, #label: {}'.format(self._total, len(np.unique(hcels_label))))
        if verbose:
            print("** mdl (Model Description Length) for describing the graph feature: {} **".format(mdls))
            # print("Island suspicious orders: {}".format(suspicious))

        return self._hcel, hcels_label, mdls
        
    def graph_node_cluster(self, node2hcel, outfn_node2lab, outfn_hcel2lab=None, 
                           strictness=4, comments="#", sep=","):
        
        outliers_marker = -1
        hcls_labs, _ = self._hcels_labeling_(self._hcel, outliers_marker, strictness)
        hcel2lab = dict(zip(map(tuple, self._hcel), hcls_labs))
        
        with open(outfn_node2lab, 'w') as ofp:
            ofp.writelines(comments + " #pt: {}, #label: {}\n".format(len(node2hcel), len(np.unique(hcls_labs))))
            for k in range(len(node2hcel)):
                ndidx, pos = node2hcel[k, 0], tuple(node2hcel[k, 1:])
                ndlab = hcel2lab.get(pos, -1)
                ofp.writelines("{}{}{}\n".format(ndidx, sep, ndlab))
            ofp.close()
        
        if outfn_hcel2lab is not None:
            np.savetxt(outfn_hcel2lab, np.vstack((self._hcel.T, hcls_labs)).T, '%i', delimiter=sep, comments=comments,
                       header='#hcel: {}, #label: {}'.format(self._total, len(np.unique(hcls_labs))))
        
    def cluster_weighted_suspicious(self, hcel_weights, strictness=4, verbose=True):
        outliers_marker = -1
        hcels_lab, hcels_prob = self._hcels_labeling_(self._hcel, outliers_marker, strictness)
        _susp_ = self._measure_suspicious_(self._hcel, np.log2(1 + self._count))

        cls2susp = dict(_susp_)
        cls2wtsusp = dict()

        hcels_lab = np.asarray(hcels_lab)
        clsid2labs = dict(zip(self.describes.keys(), range(len(self.describes.values()))))
        for clsid, desc in self.describes.items():
            lb = clsid2labs[clsid]
            lb_hcels_idx = np.arange(len(hcels_lab))[hcels_lab == lb]
            cls_wt, npts = 0, 0
            for hidx in lb_hcels_idx:
                cls_wt += hcel_weights[hidx] * self._count[hidx]
                npts += self._count[hidx]
            cls_wt /= 1.0 * npts
            cls2wtsusp[clsid] = cls2susp[clsid] * np.log(cls_wt)

        sorted_wtsusp = sorted(cls2wtsusp.items(), key=operator.itemgetter(1), reverse=True)

        if verbose:
            _suspstr_ = ", ".join(["(%d, %.2f)"%(d[0], d[1]) for d in _susp_])
            print("kl-divergence based suspicious: [{}]".format(_suspstr_))
            _wtsuspstr_ = ", ".join(["(%d, %.2f)"%(d[0], d[1]) for d in sorted_wtsusp])
            print("weighted suspicious orders: [{}]".format(_wtsuspstr_))

        return sorted_wtsusp




#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

# #  WaterLevel tree class for identifying micro-clusters / islands in
#    multi-dimensional (>=2) histogram
#  Author: wenchieh
#
#  Project: eaglemine
#      leveltree.py
#      Version:  1.0
#      Date: November 27 2017
#      Main Contact: Wenchieh Feng (wenchiehfeng.us@gmail.com)
#
#      Copyright:
#        This software is free of charge under research purposes.
#        For commercial purposes, please contact the author.
#
#      Created by @wenchieh  on <11/27/2017>


__author__ = 'wenchieh'

# sys
import operator
from itertools import product
from collections import deque

# third-party lib
import numpy as np
import scipy.ndimage.morphology as snm


class LevelTree(object):
    class Node(object):
        _DELIMITER_ = ';'

        def __init__(self, id, level, hcubes, prnt_id, child=None):
            self.id = id
            self.level = level
            self.hcubes = hcubes
            self.prnt_id = prnt_id
            self.child = child
            ## the number of pts contained in its hypercubes
            # self.mass = None

        def get_volume(self):
            return len(self.hcubes)

        def get_covers(self):
            return self.hcubes

        def __str__(self):
            # node object to string
            obj_str = "{}{}{}{}{}".format(self.id, self._DELIMITER_, self.level, self._DELIMITER_, self.prnt_id)
            childs_str = "0" if self.child is None else str(self.child).replace(', ', ',')[1:-1]
            content = ""
            for hcb in self.hcubes:
                content += str(hcb) + self._DELIMITER_
            return obj_str + self._DELIMITER_ + childs_str + self._DELIMITER_ + content[:-1]

        @staticmethod
        def load(node_str):
            if node_str is None or len(node_str) <= 0:
                return None
            tokens = node_str.strip().split(LevelTree.Node._DELIMITER_)
            node_id, level, prnt_id = int(tokens[0]), float(tokens[1]), int(tokens[2])
            child = None if tokens[3] == "0" else list(map(int, tokens[3].split(',')))
            hcubes = [eval(tokens[i]) for i in range(4, len(tokens))]

            node = LevelTree.Node(node_id, level, hcubes, prnt_id, child)
            return node

    def __init__(self):
        self.comp_tree = dict()
        self.levels = None
        self.bound = None
        self._earth = None
        self._value = None
        self._Nnnz = 0
        self.mode = 0

    #TODO optimize the search process with [scipy.ndimage.label]
    def _search_components_(self, min_level=1.0, morphology=False, sqrprobe_size=2):
        coord2val_dict = dict()
        if morphology:
            hypr_canvas = np.zeros(self.bound)
            for k in range(self._Nnnz):
                hypr_canvas[tuple(self._earth[k])] += self._value[k]
            probe = np.ones(tuple([sqrprobe_size] * hypr_canvas.ndim))
            coords = np.nonzero(snm.binary_opening(hypr_canvas, structure=probe).astype(int))
            for pos in zip(*coords):
                coord2val_dict[pos] = hypr_canvas[pos]
        else:
            coord2val_dict = dict(zip(map(tuple, self._earth), self._value))

        connect_comps = list()
        coords = list(coord2val_dict.keys())
        while len(coords) > 0:
            pos = coords.pop()
            if coord2val_dict[pos] < min_level: continue

            component = list()
            queue = deque()
            queue.append(pos)
            while len(queue) > 0:
                cur_pos = queue.popleft()
                component.append(cur_pos)
                if cur_pos in coords:
                    coords.remove(cur_pos)

                coord_nbr_rngs = list()
                for mod in range(len(cur_pos)):
                    low, up = np.max([cur_pos[mod] - 1, 0]), np.min([cur_pos[mod] + 1, self.bound[mod]])
                    if low == up:
                        coord_nbr_rngs.append([low])
                    else:
                        coord_nbr_rngs.append(range(low, up + 1))

                for nbr in product(*coord_nbr_rngs):
                    if (pos != nbr) and (nbr in coords) and (coord2val_dict[nbr] >= min_level):
                        if nbr not in queue:
                            queue.append(nbr)
            if len(component) > 0:
                connect_comps.append(component)

        return connect_comps

    def get_nodesid_atlevel(self, level):
        nodesid = list()
        for id in self.comp_tree.keys():
            node = self.comp_tree[id]
            if node.level == level:
                nodesid.append(node.id)
            if node.level > level:
                break
        return nodesid

    def get_node(self, node_id):
        node = None
        if node_id in self.comp_tree.keys():
            node = self.comp_tree.get(node_id)
        return node

    def get_nodes(self):
        return self.comp_tree.keys()

    def get_leaves(self):
        leaves = []
        for id, node in self.comp_tree.items():
            if node.child is None:
                leaves.append(id)
        return leaves

    def build_level_tree(self, nnz_indices, nnz_elems, min_level, max_level=None, step=1.0,
                         min_hcubes=4, morphology=True, verbose=False, outfn=None):
        self.levels = list()
        self.bound = list(map(int, np.max(nnz_indices, axis=0) + 1))
        self._earth = np.asarray(nnz_indices, int)
        self._value = np.asarray(nnz_elems)
        self._Nnnz = len(nnz_indices)
        self.mode = len(self.bound)

        tiny_blobs = []
        unique_values = np.unique(self._value)
        pos2val = dict(zip(map(tuple, nnz_indices), self._value))

        if (max_level is None) or (max_level < min_level) or (max_level > np.max(self._value)):
            max_level = np.max(self._value)

        nodeid = 0
        for level in np.arange(min_level, max_level, step):
            if (level > min_level) and (2**level - 2**(self.levels[-1]) < 1.0):
                continue

            if level > min_level:
                inter_npts = np.sum((unique_values <= level) & (unique_values > self.levels[-1]))
                if inter_npts <= 0:
                    continue

            connect_comps = self._search_components_(level, morphology)
            Ncomps = len(connect_comps)
            if Ncomps == 0 or (Ncomps == 1 and len(connect_comps[0]) < min_hcubes):
                if verbose:
                    print("Info: maximum tree level: {}".format(level - step))
                break

            inserted = False
            prnt_id = -1
            if level == min_level:
                for comp in connect_comps:
                    if len(comp) <= min_hcubes:
                        tiny_blobs.extend(comp)
                        if verbose:
                            for p in comp:
                                print("{}:{}".format(p, int(2**pos2val[p]-1)))
                        continue

                    inserted = True
                    node = self.Node(nodeid, level, comp, prnt_id)
                    self.comp_tree[nodeid] = node
                    nodeid += 1
            else:
                uplevel_ids = self.get_nodesid_atlevel(self.levels[-1])
                for comp in connect_comps:
                    if len(comp) <= min_hcubes:
                        tiny_blobs.extend(comp)
                        # if verbose:
                        #     for p in comp:
                        #         print("{}:{}".format(p, int(2**pos2val[p]-1)))
                        continue
                    if len(uplevel_ids) > 1:
                        shares = [len(set(comp) & set(self.comp_tree[pid].hcubes)) for pid in uplevel_ids]
                        prnt_id = uplevel_ids[np.argmax(shares)]
                    else:
                        prnt_id = uplevel_ids[0]

                    inserted = True
                    node = self.Node(nodeid, level, comp, prnt_id)
                    self.comp_tree[nodeid] = node
                    if self.comp_tree[prnt_id].child is None:
                        self.comp_tree[prnt_id].child = list()
                    self.comp_tree[prnt_id].child.append(nodeid)
                    nodeid += 1

            if inserted is True:
                self.levels.append(level)
            else:
                if verbose:
                    print("Info: maximum tree level: {}".format(level - step))
                break

        if outfn is not None:
            if len(tiny_blobs) > 0:
                blob2cnt = dict()
                for blob in tiny_blobs:
                    blob2cnt[blob] = int(2**pos2val[blob]) - 1
                blob_pos2cnt = np.zeros((len(blob2cnt), self.mode + 1),int)
                blob_pos2cnt[:, :self.mode] = np.asarray(list(blob2cnt.keys()), int)
                blob_pos2cnt[:, self.mode] = np.asarray(list(blob2cnt.values()), int)
                np.savetxt(outfn, blob_pos2cnt, '%d', ',')

        if verbose is True:
            print("Info: Level-Tree build done!")

    def dump(self, verbose=True):
        print("Level tree basic information:")
        print("#node: {}, #level: {}".format(len(self.comp_tree), len(self.levels)))
        print("(node id): {@level, parent node, #childs, #elements}")
        nodesid = list(sorted(self.comp_tree.keys(), reverse=True))

        while len(nodesid) > 0:
            stack = deque()
            stack.append(nodesid.pop())
            while len(stack) > 0:
                node_id = stack.pop()
                if node_id in nodesid:
                    nodesid.remove(node_id)
                node = self.comp_tree.get(node_id)
                level_idx = self.levels.index(node.level)
                info = "{}, {}, {}, {}".format(node.level, node.prnt_id, node.child, node.get_volume())
                print("    " * level_idx + "|---- ({}): {} |-".format(node.id, info))
                if node.child is not None:
                    stack.extend(sorted(node.child, reverse=True))

        if verbose:
            print("dump done!")

    def __str__(self):
        infns = "Level-ree info:\n"
        infns += "mode:{}, #level:{}, #hyper-cube:{}\n".format(self.mode, len(self.levels), self._Nnnz)
        infns += "(multi-)histogram bound: {} \n".format(self.bound)
        infns += "tree levels: " + ','.join(map(str, ["%.1f"%lv for lv in self.levels])) + "\n"
        infns += "#node:{}\n".format(len(self.comp_tree))
        infns += "\t id, @level, parent, child, #element\n"
        for id, node in self.comp_tree.items():
            infns += "\t {}, {}, {}, {}, {}\n".format(node.id, node.level, node.prnt_id, node.child, node.get_volume())

        return infns

    def tree_contract(self, verbose=False):
        # level-tree nodes contract
        queue_nodesid = deque(sorted(self.comp_tree.keys(), reverse=True))
        while len(queue_nodesid) > 0:
            cur_id = queue_nodesid.popleft()
            node = self.comp_tree.get(cur_id)
            prnt_id = node.prnt_id
            if prnt_id != -1:  # tree node collapse for internal nodes
                if len(self.comp_tree.get(prnt_id).child) <= 1:
                    self.comp_tree[prnt_id].child = node.child
                    if node.child is not None:
                        for child_id in node.child:
                            self.comp_tree[child_id].prnt_id = prnt_id
                    del self.comp_tree[cur_id]

        # reserve remain levels
        remain_levels = set()
        for id, node in self.comp_tree.items():
            remain_levels.add(node.level)
        self.levels = sorted(list(remain_levels))

        if verbose:
            print("Info: Level-tree contract done!")

    def tree_prune(self, alpha=0.5, verbose=False):
        # level-tree prune and smooth tiny nodes
        pos2val = dict(zip(map(tuple, self._earth), self._value))
        node2wts = dict()

        nodesid = deque(self.get_nodesid_atlevel(self.levels[0]))
        while len(nodesid) > 0:
            node = self.comp_tree[nodesid.popleft()]
            if node.child is not None:
                # N_nd_hcbs = node.get_volume()
                if node.id not in node2wts:
                    node2wts[node.id] = np.sum([2**pos2val.get(tuple(p)) for p in node.get_covers()])
                # N_child_hcbs = 0
                N_chwts = 0
                for child_id in node.child:
                    if child_id not in node2wts:
                        node2wts[child_id] = np.sum([2**pos2val.get(tuple(p))
                                                     for p in self.comp_tree[child_id].get_covers()])
                    N_chwts += node2wts.get(child_id)
                    # N_child_hcbs += self.comp_tree[child_id].get_volume()
                # if N_child_hcbs < alpha * N_nd_hcbs:
                if N_chwts <  alpha * node2wts.get(node.id):
                    self._delete_node_(node.id)
                else:
                    nodesid.extend(node.child)

        if verbose is True:
            print("Info: Level-tree prune done!")

    def _delete_node_(self, node_id):
        node = self.comp_tree.get(node_id)
        del_childids = node.child
        if del_childids is not None:
            while len(del_childids):
                del_node = self.comp_tree.get(del_childids.pop())
                if del_node.child is not None:
                    del_childids.extend(del_node.child)
                del self.comp_tree[del_node.id]
        node.child = None

    def _get_child_ids_(self, prnt_id):
        child_ids = list()
        if prnt_id == -1:
            child_ids = self.get_nodesid_atlevel(self.levels[0])
        elif prnt_id in self.comp_tree.keys():
            child_ids = self.comp_tree.get(prnt_id).child
        return child_ids

    def _get_node_margins_(self, hcube, covers):
        hcube_margins = set()
        coord_rngnbrs = []
        for mod in range(len(hcube)):
            low, up = np.max([hcube[mod] - 1, 0]), np.min([hcube[mod] + 1, self.bound[mod]])
            if low == up:
                coord_rngnbrs.append([low])
            else:
                coord_rngnbrs.append(range(low, up+1))

        for neighbor_hcb in product(*coord_rngnbrs):
            if neighbor_hcb != hcube:
                non_neg = np.sum(neighbor_hcb) == np.sum(np.abs(neighbor_hcb))  # valid hyper-cube (non-negative pos)
                if non_neg and (neighbor_hcb not in covers):
                    hcube_margins.add(neighbor_hcb)
        return hcube_margins

    def _get_comp_margins(self, node_id):
        comp_margins = set()
        if node_id in self.comp_tree.keys():
            node = self.get_node(node_id)
            for cb in node.hcubes:
                hcube_margins = self._get_node_margins_(cb, node.hcubes)
                if len(hcube_margins) > 0:
                    comp_margins = comp_margins.union(hcube_margins)
        return comp_margins

    def get_heavynodes(self):
        heavynodes2level = dict()

        parent = -1
        while True:
            node_ids = self._get_child_ids_(parent)
            volumes = [self.get_node(id).get_volume() for id in node_ids]
            max_id = node_ids[np.argmax(volumes)]
            max_node = self.get_node(max_id)
            heavynodes2level[max_id] = max_node.level
            if max_node.child is None:
                break
            parent = max_id
        return heavynodes2level

    def tree_node_expand(self, verbose=False):
        # expand islands over same levels simultaneously
        # scope: peripheral(self-uncovered) but parent-covered cells layer-diffusing.
        # expanding order: from smallest to largest (island size), top to bottom;
        # stop criteria:
        #      1. the expanded size is same as the size it was
        #      2. has overlapping among components' expending during the process.
        #      3. has no more hcubes can be included. (no update)
        expand_queue = deque()
        expand_queue.append(-1)

        while len(expand_queue) > 0:
            prnt_id = expand_queue.popleft()
            child_ids = self._get_child_ids_(prnt_id)

            if child_ids is not None:
                if prnt_id == -1:
                    remain_hcbs = set(map(tuple, self._earth))
                else:
                    prnt_covers = self.comp_tree.get(prnt_id).get_covers()
                    remain_hcbs = set(map(tuple, prnt_covers))

                child_cores = dict()
                child_size = dict()
                for nodeid in child_ids:
                    node = self.comp_tree.get(nodeid)
                    child_cores[nodeid] = set(node.get_covers())
                    child_size[nodeid] = node.get_volume()
                    remain_hcbs = remain_hcbs.difference(child_cores[nodeid])

                sorted_ids = np.asarray(sorted(child_size.items(), key=operator.itemgetter(1)))[:, 0]
                node_expands, node_margins = dict(), dict()
                tol_expands = set()
                no_overlap = True

                # non-overlap expanding
                while no_overlap:
                    updated = False
                    kexp_size = dict()
                    for nodeid in sorted_ids:
                        if (nodeid in node_expands) and  (kexp_size.get(nodeid, 0) >= 0) and  \
                                (len(node_expands[nodeid]) >= child_size[nodeid]):
                            continue
                        else:
                            # Expend current child node
                            # 1. detect all peripheral hyper cubes as the margins for the node
                            if nodeid not in node_margins:
                                node_margins[nodeid] = self._get_comp_margins(nodeid)

                            expand_hcbs = set(node_margins[nodeid]).intersection(remain_hcbs)

                            # stop criteria check: (whether overlap with other expands so no further expansion)
                            if len(expand_hcbs.intersection(tol_expands)) > 0:
                                expand_hcbs = expand_hcbs - tol_expands
                                no_overlap = False

                            if len(expand_hcbs) < kexp_size.get(nodeid, 0):
                                kexp_size[nodeid] = -1
                                continue

                            if len(expand_hcbs) > 0:
                                # update total expand set
                                tol_expands = tol_expands.union(expand_hcbs)
                                if nodeid not in node_expands:  # # haven't been expanded so far
                                    node_expands[nodeid] = set()
                                node_expands[nodeid] = node_expands[nodeid].union(expand_hcbs)
                                kexp_size[nodeid] = len(expand_hcbs)
                                # at least one node has been updated (expanded)
                                updated = True

                                if no_overlap:
                                    # only no-overlap holds, the updating for margins of current node
                                    # will happen in next iteration
                                    # update current node margins
                                    expanded_covers = node_expands[nodeid].union(child_cores[nodeid])
                                    nextiter_margins = set()
                                    for cb in expand_hcbs:
                                        nextiter_margins = nextiter_margins.union(self._get_node_margins_(cb, expanded_covers))
                                    node_margins[nodeid] = nextiter_margins.intersection(remain_hcbs)
                                    # end if can be expand
                                    # end (else) if the island can be expand further.
                    # end for (all child nodes)
                    if updated is False:  # no expansion occur in this step.
                        break
                # end while for non-overlap expanding

                # Execute update tree node with expandable covers
                for nodeid in child_ids:
                    if nodeid not in node_expands: continue
                    self.comp_tree[nodeid].hcubes = list(node_expands[nodeid].union(child_cores[nodeid]))
                    expand_queue.append(nodeid)
            # end if (expand for internal tree node)
        # end while (expand each nodes with constraints if possible)

        if verbose is True:
            print("Info: Level-tree expand done!")

    def save_leveltree(self, outfn, sep=',', verbose=True):
        with open(outfn, 'w') as ofp:
            header = '# mode, bound, levels, #non-zeros, #tree-node, earth_value, tree_nodes\n'
            ofp.writelines(header)
            ofp.writelines(str(self.mode) + '\n')
            ofp.writelines(str(self.bound)[1:-1].replace(', ', sep) + '\n')
            ofp.writelines(sep.join(['%.1f' % lv for lv in self.levels]) + '\n')
            ofp.writelines(str(self._Nnnz) + '\n')
            ofp.writelines(str(len(self.comp_tree)) + '\n')
            for k in range(self._Nnnz):
                ofp.writelines(str(list(self._earth[k]))[1:-1].strip().replace(', ', sep) +
                               sep + str(self._value[k]) + '\n')

            for nodeid, comp in self.comp_tree.items():
                ofp.writelines(str(comp) + '\n')
            ofp.close()

        if verbose:
            print("Level-tree serialization done!")

    def load_leveltree(self, infn, sep=',', verbose=True):
        with open(infn, 'r') as ifp:
            ifp.readline()     # header line
            self.mode = int(ifp.readline().strip())
            self.bound = list(map(int, ifp.readline().strip().split(sep)))
            self.levels = list(map(float, ifp.readline().strip().split(sep)))
            self._Nnnz = int(ifp.readline().strip())
            Nnodes = int(ifp.readline().strip())
            self.mode = len(self.bound)
            self._earth = np.zeros((self._Nnnz, self.mode), int)
            self._value = np.zeros(self._Nnnz, float)
            for k in range(self._Nnnz):
                tokens = ifp.readline().strip().split(sep)
                self._earth[k] = np.array(list(map(int, tokens[:self.mode])), int)
                self._value[k] = float(tokens[self.mode])

            self.comp_tree = dict()
            for n in range(Nnodes):
                node = self.Node.load(ifp.readline().strip())
                if node is not None:
                    self.comp_tree[node.id] = node
            ifp.close()

        if verbose:
            print("Level-tree load done!")


# sys
import os
import sys
import time
import argparse
import gc

# third-part libs
import numpy as np
import scipy.sparse.linalg as linalg
import scipy.sparse

# project
from spartan.model.fraudar.greedy import logWeightedAveDegree, sqrtWeightedAveDegree, aveDegree, fast_greedy_decreasing_monosym
from .._model import DMmodel


class Specgreedy(DMmodel):
    def __init__(self, data_mat):
        self.data = data_mat

    def run(self, out_path = "./", file_name = "specgreedy", bipartite=False, T=5, delete_type="edge", **kwargs):
        Mcur = self.data.to_scipy().tocsr()
        res = []
        t = 0
        print("Specgreedy for find %d dense subgraphs." % T)
        while (t < T):
            print("\nRunning for %d-th subgraph" % t)
            if bipartite:
                list_row, list_col, score = self.run_bip(Mcur, **kwargs)
            else:
                list_row, list_col, score = self.run_undi(Mcur, **kwargs)
            res.append((list_row, list_col, score))

            np.savetxt("%s_%s.rows" % (out_path + file_name, t), np.array(list_row).reshape(-1, 1), fmt='%d')
            np.savetxt("%s_%s.cols" % (out_path + file_name, t), np.array(list_col).reshape(-1, 1), fmt='%d')
            print("score obtained is ", score)

            if (len(list_row) == 0 or len(list_col) == 0):
                print("Finding subgraph failed. Specgreedy stopped.")
                break

            t += 1
            gc.collect()

            if (t >= T):
                break
            
            tic = time.time()
            if delete_type == "edge":
                ## only delete inner connections
                (rs, cs) = Mcur.nonzero() # (u, v)
                rowSet = set(list_row)
                colSet = set(list_col)
                for i in range(len(rs)):
                    if rs[i] in rowSet and cs[i] in colSet:
                        Mcur[rs[i], cs[i]] = 0
            elif delete_type == "node":
                diag = scipy.sparse.eye(Mcur.shape[0]).tolil()
                for r in list_row:
                    diag[r, r] = 0
                diag = diag.tocsr()
                Mcur = diag.dot(Mcur)

                diag = scipy.sparse.eye(Mcur.shape[1]).tolil()
                for c in list_col:
                    diag[c, c] = 0
                diag = diag.tocsr()
                Mcur = Mcur.dot(diag)
            else:
                raise ValueError("Invalid argument delete_type. Please set 'edge' or 'node'")
            toc = time.time()
            print("Remove current block: %ss" % (toc-tic))

        return res
            

    def run_undi(self, sm, weighted = True, topk = 5, scale = 1.0):
        # outfn = output_path
        w_g = weighted
        
        # print("## Dataset: {}".format(infn[infn.rfind('/')+1:]))
        greedy_fun = fast_greedy_decreasing_monosym

        t0 = time.time()
        ms, ns = sm.shape
        n = max(ms, ns)
        sm -= sps.diags(sm.diagonal())
        es = sm.sum()

        if (abs(sm-sm.T)>1e-10).nnz > 0:
            sm += sm.T
        if not w_g:
            print("max edge weight: {}".format(sm.max()))
            sm = sm > 0
            sm = sm.astype('int')

        print("load graph @ {}s".format(time.time() - t0))
        print("graph: #node: {}, #edge: {}, # es: {}".format((ms, ns), es, sm.sum()))
        print("matrix max: {}, min: {}, shape: {}\n".format(sm.max(), sm.min(), (n ,n)))
        print("Finding subgraph with top k singular values:", topk)

        orgnds, cans = None, None
        opt_density = 0.0
        opt_k = -1

        k = 0
        decom_n = 0

        start = 3
        step = 3
        isbreak = False
        t1 = time.time()
        while k < topk:
            print("\nComputing top{} singular vectors and values for efficency".format(start + decom_n * step))
            U, S, V = linalg.svds(sm.asfptype(), k=start + decom_n * step, which='LM', tol=1e-4)
            U, S, V = U[:, ::-1], S[::-1], V.T[:, ::-1]
            print("lambdas:", S)
            kth  = k
            while kth < start + decom_n * step - 1 and kth < topk:
                if abs(max(U[:, kth])) < abs(min(U[:, kth])): U[:, kth] *= -1
                if abs(max(V[:, kth])) < abs(min(V[:, kth])): V[:, kth] *= -1
                row_cans = list(np.where(U[:, kth] >= 1.0 / np.sqrt(ms))[0])
                # col_cans = list(np.where(V[:, kth] >= 1.0 / np.sqrt(ns))[0])
                col_cans = row_cans
                if len(row_cans) <= 1 or len(col_cans) <= 1:
                    print("SKIP: candidates sizes are too small: {}".format((len(row_cans), len(col_cans))))
                    kth += 1
                    k += 1
                    continue
                sm_part = sm[row_cans, :][:, col_cans]
                # print("{}, size: {}".format(kth, sm_part.shape))
                row_ids, col_ids, avgsc_part = greedy_fun(sm_part)
                kth += 1
                k += 1
                gc.collect()

                print("k_cur:{}, size: {}, density: {}.  @ {}s\n".format(kth, (len(row_ids), len(col_ids)),  avgsc_part, time.time() - t1))
                if avgsc_part > opt_density:
                    opt_k, opt_density = kth, avgsc_part
                    sm_pms = max(len(row_cans), len(col_cans))
                    cans = row_cans
                    fin_pms = len(nds_res)
                    print("Update. svd init shape (candidates size): {}".format((sm_pms, sm_pms)))
                    print("Update. size: {}, score: {}\n".format((fin_pms, fin_pms), avgsc_part))
                    nd_idx = dict(zip(range(sm_pms), sorted(cans)))
                    orgnds = [nd_idx[id] for id in nds_res]

                if 2.0*opt_density >= S[kth]: # kth < topk and
                    print("Early Stopped. k_cur: {},  optimal density: {}, lambda_k: {}".format(kth, opt_density, S[kth]))
                    isbreak = True
                    break
            if isbreak:
                break
            decom_n += 1


        if orgnds is not None:
            print("\noptimals: k:{}, size:{}, density:{}".format(opt_k, fin_pms, opt_density))
            print("total time @ {}s".format(time.time() - t1))
            return orgnds, orgnds, opt_density
        else:
            print("No dense subgraphs found.")
            return [], [], 0

    def run_bip(self, sm, weighted = True, topk = 5, alpha = 5.0, scale = 1.0, col_wt = "even", maxsize = -1):
        # outfn = output_path
        w_g = weighted

        #alpha = 1.0
        greedy_func = None
        if col_wt == 'even':
            greedy_func = logWeightedAveDegree
        elif args.col_wt == 'sqrt':
            greedy_func = sqrtWeightedAveDegree
        else:
            greedy_func = aveDegree

        t0 = time.time()
        ms, ns = sm.shape
        if not w_g:
            print("max edge weight: {}".format(sm.max()))
            sm = sm > 0
            sm = sm.astype('int')
        es = sm.sum()
        print("load graph time @ {}s".format(time.time() - t0))
        print("graph: #node: {},  #edge: {}".format((ms, ns), es))
        print("matrix max: {}, min: {}, shape: {}\n".format(sm.max(), sm.min(), sm.shape))
        print("Finding subgraph with top k singular values:", topk)

        opt_k = -1
        opt_density = 0.0
        orgnds, cans = None, None
        fin_pms, fin_pns = 0, 0

        k = 0
        decom_n = 0

        start = 3
        step = 3
        isbreak = False
        orgnds = None

        t1 = time.time()  
        while k < topk:
            print("\nComputing top{} singular vectors and values for efficency".format(start + decom_n * step))
            U, S, V = linalg.svds(sm.asfptype(), k=start + decom_n * step, which='LM', tol=1e-4)
            U, S, V = U[:, ::-1], S[::-1], V.T[:, ::-1]
            print("lambdas:", S)
            kth  = k
            while kth < start + decom_n * step - 1 and kth < topk:
                if abs(max(U[:, kth])) < abs(min(U[:, kth])):
                    U[:, kth] *= -1
                if abs(max(V[:, kth])) < abs(min(V[:, kth])):
                    V[:, kth] *= -1
                row_cans = list(np.where(U[:, kth] >= 1.0 / np.sqrt(ms))[0])
                col_cans = list(np.where(V[:, kth] >= 1.0 / np.sqrt(ns))[0])
                if len(row_cans) <= 1 or len(col_cans) <= 1:
                    print("SKIP: candidates sizes are too small: {}".format((len(row_cans), len(col_cans))))
                    kth += 1
                    k += 1
                    continue
                sm_part = sm[row_cans, :][:, col_cans]
                row_ids, col_ids, avgsc_gs = greedy_func(sm_part, maxsize)
                
                kth += 1
                k += 1

                print("k_cur: {} size: {}, density: {}  @ {}s".format(kth, (len(row_ids), len(col_ids)), 
                                                                    avgsc_gs, time.time() - t1))
                if avgsc_gs > opt_density:
                    opt_k, opt_density = kth, avgsc_gs
                    (sm_pms, sm_pns) = sm_part.shape
                    fin_pms, fin_pns = len(row_ids), len(col_ids)
                    print("Update. svd tops shape (candidates size): {}".format((sm_pms, sm_pns)))
                    print("Update. size: {}, score: {}\n".format((fin_pms, fin_pns), avgsc_gs))

                    row_idx = dict(zip(range(sm_pms), sorted(row_cans)))
                    col_idx = dict(zip(range(sm_pns), sorted(col_cans)))
                    org_rownds = [row_idx[id] for id in row_ids]
                    org_colnds = [col_idx[id] for id in col_ids]
                    cans = [row_cans, col_cans]
                    orgnds = [org_rownds, org_colnds]
                    
                if 2.0 * opt_density >= S[kth]: # kth < topk and
                    print("Early Stopped. k_cur: {},  optimal density: {}, lambda_k: {}".format(kth, opt_density, S[kth]))
                    isbreak = True
                    break
            if isbreak:
                break
            decom_n += 1
        
        if orgnds is not None:
            print("\noptimal k: {}, density: {}".format(opt_k, opt_density))    
            print("total time @ {}s".format(time.time() - t1))
            return orgnds[0], orgnds[1], opt_density
        else:
            print("No dense subgraphs found.")
            return [], [], 0

    
    def anomaly_detection(self):
        return self.run()

    def save(self, outpath):
        pass


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   summarization.py
@Desc    :   Implementation for summarization task.
'''

# here put the import lib

from . import MODEL_PATH

from ._task import Task
from enum import Enum


class Summarization(Task):
    '''Implementation for summarization task.
    '''

    def run(self, *args, **kwargs):
        '''Call summarization function of selected model.

        If not implemented, raise an exception by calling parent run.
        '''
        if "summarization" in dir(self.model):
            return self.model.summarization(*args, **kwargs)
        else:
            return super().run("summarization")

    @classmethod
    def list_models(cls):
        '''List models which can do summarization task.
        '''
        for name in SumPolicy.__members__:
            print(name)


class SumPolicy(Enum):
    '''Registration for path of models who can do summarization task.
    '''
    BeatLex = MODEL_PATH + ".beatlex"
    DPGS = MODEL_PATH + ".DPGS"
    kGrass = MODEL_PATH + ".kGS"



from biosppy.signals import ecg

from .._model import DMmodel
from . import param_default, DTensor

param_default_dict = {
    'sampling_rate': 360,
    'left_size': 120,
    'right_size': 136,
    'out_path': None
}


class RPeak(DMmodel):
    def __init__(self, time_series, *args, **para_dict):
        self.series_data = time_series.val_tensor
        self.sampling_rate = param_default(para_dict, 'sampling_rate', param_default_dict)
        self.left_size = param_default(para_dict, 'left_size', param_default_dict)
        self.right_size = param_default(para_dict, 'right_size', param_default_dict)
        self.out_path = param_default(para_dict, 'out_path', param_default_dict)
        self.length = time_series.length
        self.segments = None

    def _find_rpeaks(self):
        sig_out = ecg.ecg(signal=self.series_data[0], sampling_rate=self.sampling_rate, show=False)
        r_peaks = sig_out["rpeaks"]
        segments = []
        for r_peak in r_peaks:
            if r_peak-self.left_size >= 0 and r_peak + self.right_size < self.length:
                ts = self.series_data[:, r_peak-self.left_size:r_peak + self.right_size]
                segments.append(ts)
        import numpy as np
        self.segments = DTensor.from_numpy(np.array(segments))
        if self.out_path is not None:
            np.save(self.out_path, segments)
        return self.segments

    def run(self):
        return self._find_rpeaks()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py
@Desc    :   Import part and configuration part, including names exposed to spartan.
'''

# here put the import lib

from spartan.backend import STensor, DTensor

from .tensor import TensorData
from .tensor import TensorData, TensorStream
from .timeseries import Timeseries
from .graph import Graph

__all__ = [
    'TensorData', 'TensorStream', 'Timeseries', 'Graph'
]




from torch.utils.data import DataLoader, Dataset
import torch
from . import param_default


class MyDataSet(Dataset):
    def __init__(self, segments, labels):
        self.data = torch.Tensor(segments)
        self.labels = torch.Tensor(labels)

        self.data = self.data.transpose(1, 2)

    def __getitem__(self, index):
        return self.data[index], self.labels[index]

    def __len__(self):
        return self.data.shape[0]


def preprocess_data(data, labels=None, param=None, is_train=True):
    import numpy as np
    if labels is None:
        labels = np.zeros([data.shape[0], 1])

    dataset = MyDataSet(data, labels)
    batch_size = param_default(param, "batch_size", 64)

    data_loader = DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        shuffle=is_train,
        drop_last=is_train
    )

    return data_loader



import functools
import logging
import math
import os
import time
from collections import Counter, defaultdict

import numpy as np
import scipy.sparse as ssp
from datasketch import MinHashLSH

from . import c_MDL
from .neiminhash import NeiMinHash
from .union_find import UnionFind

from .._model import DMmodel
from spartan import STensor

logger = logging.getLogger('summarize')
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    '%(asctime)s %(filename)s %(lineno)d %(levelname)s: %(message)s')

if len(logger.handlers) < 2:
    filename = os.path.join(os.path.dirname(__file__), 'summarize.log')
    file_handler = logging.FileHandler(filename, mode='w')
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)


@functools.lru_cache(maxsize=10000)
def LN(n):
    return c_MDL.LN(n)


@functools.lru_cache(maxsize=10000)
def xlogx(x):
    if x == 0:
        return 0
    return x * math.log(x)


class DPGSummarizer(DMmodel):
    # Maximum size of each group
    C = 500

    def __init__(self, graph):
        self.sm = graph.sm.tolil()
        self.sm.setdiag(0)
        N = self.N = self.sm.shape[0]
        self.es = self.sm.nnz // 2
        self.minhashes = [None] * N
        self.deleted = set()
        self.adj = [Counter(self.sm.rows[n]) for n in range(N)]

        self.degs = np.array(self.sm.sum(axis=1)).flatten()
        self.sizes = [1] * self.N
        self.nodes_dict = dict(zip(range(N), ({n} for n in range(N))))

    def run(self, T=30, n_sign=16, min_b=3, max_b=8):
        return self._summarize(T, n_sign, min_b, max_b)

    def summarization(self, T=30, n_sign=16, min_b=3, max_b=8):
        return self._summarize(T, n_sign, min_b, max_b)

    def _merge(self, u, v):
        nodes_dict = self.nodes_dict
        sizes = self.sizes
        degs = self.degs

        nodes_dict[u] = nodes_dict[u] | nodes_dict[v]
        del nodes_dict[v]

        # Update degree and sizes
        degs[u] = degs[u] + degs[v]
        degs[v] = 0
        sizes[u] = sizes[u] + sizes[v]
        sizes[v] = 0

        uv = self.adj[u][v]
        self.adj[u].update(self.adj[v])
        self.adj[u][u] += (uv + self.adj[v][v])
        if self.adj[u][u] == 0:
            del self.adj[u][u]
        del self.adj[u][v]
        self.adj[v] = None
        for nei in self.adj[u]:
            if nei != u and nei != v and nei not in self.deleted:
                if self.adj[u][nei] != 0:
                    self.adj[nei][u] = self.adj[u][nei]
        self.deleted.add(v)

    def _update_lsh(self, params):
        for n in self.nodes_dict:
            if n in self.deleted:
                continue
            neighbors = [n_ for n_ in self.adj[n] if n_ not in self.deleted]
            if len(neighbors) == 0:
                continue
            m = self.minhashes[n]
            if m is None:
                m = NeiMinHash(num_perm=self.n_sign, hashfunc=hash, seed=1024)
            else:
                m.clear()
            m.updates(neighbors)
            self.minhashes[n] = m

        self.lsh = MinHashLSH(num_perm=self.n_sign, params=params)
        lsh = self.lsh
        with lsh.insertion_session() as session:
            for n in self.nodes_dict:
                if self.minhashes[n] is not None:
                    session.insert(str(n), self.minhashes[n])

        uf = UnionFind(self.N)
        for table in lsh.hashtables:
            d = table._dict
            for k, v in d.items():
                nodes = list(v)
                if len(nodes) <= 1:
                    continue
                n = int(nodes[0])
                for n_ in nodes[1:]:
                    uf.union(n, int(n_))
        group_dict = defaultdict(list)
        for n in self.nodes_dict:
            group_dict[uf._root(n)].append(n)
        groups = []
        for v in group_dict.values():
            while len(v) > self.C:
                groups.append(v[:self.C])
                v = v[self.C:]
            if len(v) > 0:
                groups.append(v)

        return groups

    def _merge_gain(self, u, v):
        neiu = set((n_ for n_ in self.adj[u] if n_ not in self.deleted))
        neiv = set((n_ for n_ in self.adj[v] if n_ not in self.deleted))
        du = self.degs[u]
        dv = self.degs[v]

        gain = LN(self.cnt) - LN(self.cnt-1)
        gain += 2 * (xlogx(du) + xlogx(dv) - xlogx(du+dv))
        # gain += self.N * math.log2(self.cnt / (self.cnt-1))
        gain += self.N * (LN(self.cnt) - LN(self.cnt-1))
        common_nei = 0
        for nei in neiu:
            if nei not in neiv:
                continue
            if nei == u or nei == v:
                continue
            common_nei += 1
            new_weight = self.adj[u][nei] + self.adj[v][nei]
            gain -= 2 * (xlogx(self.adj[u][nei]) + xlogx(self.adj[v][nei]))
            gain += 2 * xlogx(new_weight)
            gain += LN(self.adj[u][nei]) + LN(self.adj[v][nei])
            gain -= LN(new_weight)
        new_es = self.es - common_nei
        # Dealing with self-loop
        if u in neiu or v in neiv or u in neiv:
            new_weight = self.adj[u][u] + self.adj[v][v] + 2*self.adj[u][v]
            gain -= LN(new_weight)
            gain += xlogx(new_weight)
            if u in neiu:
                gain += LN(self.adj[u][u])
                gain -= xlogx(self.adj[u][u])
            if v in neiv:
                gain += LN(self.adj[v][v])
                gain -= xlogx(self.adj[v][v])
            if u in neiv:
                gain += LN(self.adj[u][v])
                gain -= 2 * xlogx(self.adj[u][v])
            new_es -= ((u in neiu) + (v in neiv) + (u in neiv) - 1)

        gain += LN(self.es) - LN(new_es)
        cnt = self.cnt
        gain += c_MDL.LnU(cnt*(cnt+1)//2, self.es)
        gain -= c_MDL.LnU(cnt*(cnt-1)//2, new_es)
        return gain, new_es

    def _merge_group(self, group: list):
        times = max(int(math.log2(len(group))), 1)
        merge_cnt = 0
        num_skip = 0
        gains = 0.0
        while True:
            if len(group) <= 1:
                break
            max_gain = 0
            best_pair = None
            best_new_e = 0
            for i in range(times):
                if len(group) <= 1:
                    break
                u, v = np.random.choice(group, 2)
                while u == v:
                    u, v = np.random.choice(group, 2)
                gain, new_e = self._merge_gain(u, v)
                if gain > max_gain:
                    max_gain = gain
                    best_pair = (u, v)
                    best_new_e = new_e
            if max_gain > 0 and best_pair is not None:
                u, v = best_pair
                logger.debug(f"Merge {u} and {v}, gain:{max_gain}")
                gains += max_gain
                merge_cnt += 1
                num_skip = 0
                self._merge(u, v)
                self.cnt -= 1
                self.es = best_new_e
                group.remove(v)
            else:
                num_skip += 1

            if num_skip >= times:
                break
        return merge_cnt, gains

    @staticmethod
    def _check_parameters(T, n_sign, min_b, max_b):
        if T <= 0:
            print(f"`T`({T}) should be greater than 0")
            return False
        if n_sign < 0:
            print(f"`n_sign`({n_sign}) should be greater than 0")
            return False
        if min_b < 0:
            print(f"`min_b`({min_b}) should be greater than 0")
            return False
        if max_b < 0:
            print(f"`max_b`({max_b}) should be greater than 0")
            return False
        if max_b > n_sign:
            print(f"`max_b`({max_b}) should be less or equal than `n_sign`({n_sign})")
            return False
        if min_b > max_b:
            print(f"`min_b`({min_b}) should be less or equal than `max_b`({max_b})")
            return False
        return True

    def _summarize(self, T, n_sign, min_b, max_b):
        N = self.N

        if not self._check_parameters(T, n_sign, min_b, max_b):
            print("Check parameter fails.")
            return
        self.n_sign = n_sign
        self.min_b, self.max_b = int(min_b), int(max_b)
        slope = (max_b - min_b) / (T-1)

        degs = self.degs
        degs_orig = degs.copy()
        nodes_dict = self.nodes_dict

        start_time = time.time()
        t = 0
        self.N = self.cnt = N

        self.total_gain = 0.0
        while t < T:
            t += 1
            b = int(self.min_b + (t-1) * slope)
            r = n_sign // b
            params = (b, r)
            start_time2 = time.time()
            groups = self._update_lsh(params)
            elapsed = time.time() - start_time2
            merge_cnt = 0
            for group in groups:
                tmp1, tmp2 = self._merge_group(group)
                merge_cnt += tmp1
                self.total_gain += tmp2
            logger.info(f"Merge {merge_cnt} in Iteration {t}")
            if merge_cnt == 0:
                logger.info(f"No improvment at iteration {t}, break.")
                break

        elapsed = time.time() - start_time
        logger.info(
            f"Summarize {N} nodes to {self.cnt} nodes with {self.es} edges, costs {elapsed} seconds")
        logger.info(f"Total gain: {self.total_gain}")

        rows, cols, datas = [], [], []
        for i, (n, nodes) in enumerate(nodes_dict.items()):
            assert len(nodes) != 0
            D = sum(degs_orig[n_] for n_ in nodes)
            nodes = list(nodes)
            rows.extend([i] * len(nodes))
            cols.extend(nodes)
            if D == 0:
                datas.extend([1.0 / len(nodes)] * len(nodes))
            else:
                datas.extend((degs_orig[n_] / D for n_ in nodes))

        P_ = ssp.csr_matrix(([1] * len(datas), (rows, cols)),
                            shape=(len(nodes_dict), self.N))
        sm_s = P_ @ self.sm @ P_.T

        self.nodes_dict = nodes_dict
        self.sm_s = sm_s
        return STensor.from_scipy_sparse(sm_s)



import numpy as np
from .._model import DMmodel
import scipy.sparse.linalg as slin
import numpy as np
from scipy.sparse import csc_matrix, coo_matrix, csr_matrix, lil_matrix
from spartan.model.fraudar.greedy import logWeightedAveDegree

class Fraudar(DMmodel):
    def __init__(self, data_mat):
        self.data = data_mat

    def run(self, out_path = "./", file_name = "out", k = 1, maxsize = -1):
        sparse_matrix = self.data.to_scipy()
        sparse_matrix = sparse_matrix.asfptype()
        Mcur = sparse_matrix.copy().tolil()
        res = []
        t = 0
        while (t < k):
            set_row, set_col, score = logWeightedAveDegree(Mcur, maxsize = maxsize)
            list_row, list_col = list(set_row), list(set_col)
            print("Fraudar iter %s finished." % t)
            print("n_row", len(list_row), "n_col:", len(list_col))
            print("score obtained is ", score)

            np.savetxt("%s_%s.rows" % (out_path + file_name, t), np.array(list_row).reshape(-1, 1), fmt='%d')
            np.savetxt("%s_%s.cols" % (out_path + file_name, t), np.array(list_col).reshape(-1, 1), fmt='%d')
            
            t += 1

            if (t >= k):
                break
            
            (rs, cs) = Mcur.nonzero() # (u, v)
            ## only delete inner connections
            rowSet = set(list_row)
            colSet = set(list_col)
            for i in range(len(rs)):
                if rs[i] in rowSet and cs[i] in colSet:
                    Mcur[rs[i], cs[i]] = 0
    
    def anomaly_detection(self):
        return self.run()

    def save(self, outpath):
        pass



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

#  Project: eaglemine
#      norm_extras.py
#           some useful sub-routines for Aderson-Darling Test, multi-variate normal CDF
#      Version:  1.0
#      Goal: Routine scripts
#      Created by @wenchieh  on <11/30/2017>


__author__ = 'wenchieh'

# third-party lib
import numpy as np
from scipy.stats import mvn
from scipy.stats import distributions
from collections import namedtuple
from statsmodels.stats.weightstats import DescrStatsW

# Values from Stephens, M A, "EDF Statistics for Goodness of Fit and
#             Some Comparisons", Journal of the American Statistical
#             Association, Vol. 69, Issue 347, Sept. 1974, pp 730-737
_Avals_norm = np.array([0.576, 0.656, 0.787, 0.918, 1.092])

AndersonResult = namedtuple('AndersonResult', ('statistic',
                                               'critical_values',
                                               'significance_level'))

INFORMCODE = {0: 'normal completion with ERROR < EPS',
              1: '''completion with ERROR > EPS and MAXPTS function values used;
                    increase MAXPTS to decrease ERROR;''',
              2: 'N > 500 or N < 1'}


def norm_anderson(x, weights=None):
    """
    Anderson-Darling test for data coming from a particular distribution

    The Anderson-Darling test is a modification of the Kolmogorov-
    Smirnov test `kstest` for the null hypothesis that a sample is
    drawn from a population that follows a particular distribution.
    For the Anderson-Darling test, the critical values depend on
    which distribution is being tested against.  This function works
    for normal, exponential distributions.

    Parameters
    ----------
    x : array_like
        array of sample data
    weight: array_like optional (default = None)
        array of weight for each sample data

    Returns
    -------
    statistic : float
        The Anderson-Darling test statistic
    critical_values : list
        The critical values for this distribution
    significance_level : list
        The significance levels for the corresponding critical values
        in percents.  The function returns critical values for a
        differing set of significance levels depending on the
        distribution that is being tested against.

    Notes
    -----
    Critical values provided are for the following significance levels:

    normal/exponential
        15%, 10%, 5%, 2.5%, 1%

    If A2 is larger than these critical values then for the corresponding
    significance level, the null hypothesis that the data come from the
    chosen distribution can be rejected.

    References
    ----------
    .. [1] http://www.itl.nist.gov/div898/handbook/prc/section2/prc213.htm
    .. [2] Stephens, M. A. (1974). EDF Statistics for Goodness of Fit and
           Some Comparisons, Journal of the American Statistical Association,
           Vol. 69, pp. 730-737.
    .. [3] Stephens, M. A. (1976). Asymptotic Results for Goodness-of-Fit
           Statistics with Unknown Parameters, Annals of Statistics, Vol. 4,
           pp. 357-369.
    .. [4] Stephens, M. A. (1977). Goodness of Fit for the Extreme Value
           Distribution, Biometrika, Vol. 64, pp. 583-588.
    .. [5] Stephens, M. A. (1977). Goodness of Fit with Special Reference
           to Tests for Exponentiality , Technical Report No. 262,
           Department of Statistics, Stanford University, Stanford, CA.
    .. [6] Stephens, M. A. (1979). Tests of Fit for the Logistic Distribution
           Based on the Empirical Distribution Function, Biometrika, Vol. 66,
           pp. 591-595.
    """
    N = len(x)
    x = np.asarray(x)
    if weights is None:
        weights = np.ones(N)
    else:
        if len(weights) != N:
            raise ValueError("Invalid weights; the weights must "
                             "have the same shape as x or None.")
        weights = np.asarray(weights)

    sorted_index = np.argsort(x)
    x, weights = x[sorted_index], weights[sorted_index]

    xw = DescrStatsW(x, weights=weights, ddof=1)
    xbar = xw.mean
    s = xw.std

    w = (x - xbar) / s
    logcdf = distributions.norm.logcdf(w)
    logsf = distributions.norm.logsf(w)

    expand_lgcdf, expand_lgsf = [], []
    x_expand = []
    for i in range(len(x)):
        expand_lgcdf.extend([logcdf[i]]*weights[i])
        expand_lgsf.extend([logsf[i]]*weights[i])
        x_expand.extend([x[i]] * weights[i])

    print(len(x_expand))
    expand_lgcdf, expand_lgsf = np.array(expand_lgcdf), np.array(expand_lgsf)
    N_w = int(np.sum(weights))
    i = np.arange(1, N_w + 1)

    A2 = -N_w - np.sum((2*i - 1.0) / N_w * (expand_lgcdf + expand_lgsf[::-1]), axis=0)
    sig = np.array([15, 10, 5, 2.5, 1])
    critical = np.around(_Avals_norm / (1.0 + 4.0/N_w - 25.0/N_w/N_w), 3)

    return AndersonResult(A2, critical, sig)


def _mvstdnormcdf(lower, upper, corrcoef, **kwds):
    """
    standardized multivariate normal cumulative distribution function
    This is a wrapper for scipy.stats.kde.mvn.mvndst which calculates
    a rectangular integral over a standardized multivariate normal
    distribution.

    This function assumes standardized scale, that is the variance in each dimension
    is one, but correlation can be arbitrary, covariance = correlation matrix
    Parameters
    ----------
    lower, upper : array_like, 1d
       lower and upper integration limits with length equal to the number
       of dimensions of the multivariate normal distribution. It can contain
       -np.inf or np.inf for open integration intervals
    corrcoef : float or array_like
       specifies correlation matrix in one of three ways, see notes
    optional keyword parameters to influence integration
        * maxpts : int, maximum number of function values allowed. This
             parameter can be used to limit the time. A sensible
             strategy is to start with `maxpts` = 1000*N, and then
             increase `maxpts` if ERROR is too large.
        * abseps : float absolute error tolerance.
        * releps : float relative error tolerance.
    Returns
    -------
    cdfvalue : float
        value of the integral
    Notes
    -----
    The correlation matrix corrcoef can be given in 3 different ways
    If the multivariate normal is two-dimensional than only the
    correlation coefficient needs to be provided.
    For general dimension the correlation matrix can be provided either
    as a one-dimensional array of the upper triangular correlation
    coefficients stacked by rows, or as full square correlation matrix
    See Also
    --------
    mvnormcdf : cdf of multivariate normal distribution without
        standardization
    :return:
    """
    n = len(lower)
    lower = np.array(lower)
    upper = np.array(upper)
    correl = np.zeros(n * (n - 1) // 2)
    corrcoef = np.array(corrcoef)

    if (lower.ndim != 1) or (upper.ndim != 1):
        raise ValueError("Error: can handle only 1D bounds")
    if len(upper) != n:
        raise ValueError("Error: bounds have different lengths")
    if n == 2 and corrcoef.size == 1:
        correl = corrcoef
        # print('case scalar rho', n)
    elif corrcoef.ndim == 1 and len(corrcoef) == n * (n - 1) / 2.0:
        # print('case flat corr', corrcoeff.shape)
        correl = corrcoef
    elif corrcoef.shape == (n, n):
        # print('case square corr',  correl.shape)
        correl = corrcoef[np.tril_indices(n, -1)]
    else:
        raise ValueError("Error: corrcoef has incorrect dimension")

    if not 'maxpts' in kwds:
        if n > 2:
            kwds['maxpts'] = 10000 * n

    lowinf = np.isneginf(lower)
    uppinf = np.isposinf(upper)
    infin = 2.0 * np.ones(n)

    np.putmask(infin, lowinf, 0)  # infin.putmask(0,lowinf)
    np.putmask(infin, uppinf, 1)  # infin.putmask(1,uppinf)
    # this has to be last
    np.putmask(infin, lowinf * uppinf, -1)

    ##    #remove infs
    ##    np.putmask(lower,lowinf,-100)# infin.putmask(0,lowinf)
    ##    np.putmask(upper,uppinf,100) #infin.putmask(1,uppinf)

    error, cdfvalue, inform = mvn.mvndst(lower, upper, infin, correl, **kwds)
    if inform:
        print("Error Something wrong. {}: {}".format(INFORMCODE[inform], error))
    return cdfvalue


def mvnormcdf(upper, mu, cov, lower=None, **kwds):
    '''multivariate normal cumulative distribution function
    This is a wrapper for scipy.stats.kde.mvn.mvndst which calculates
    a rectangular integral over a multivariate normal distribution.

    Parameters
    ----------
    lower, upper : array_like, 1d
       lower and upper integration limits with length equal to the number
       of dimensions of the multivariate normal distribution. It can contain
       -np.inf or np.inf for open integration intervals
    mu : array_lik, 1d
       list or array of means
    cov : array_like, 2d
       specifies covariance matrix
    optional keyword parameters to influence integration
        * maxpts : int, maximum number of function values allowed. This
             parameter can be used to limit the time. A sensible
             strategy is to start with `maxpts` = 1000*N, and then
             increase `maxpts` if ERROR is too large.
        * abseps : float absolute error tolerance.
        * releps : float relative error tolerance.
    Returns
    -------
    cdfvalue : float
        value of the integral
    Notes
    -----
    This function normalizes the location and scale of the multivariate
    normal distribution and then uses `mvstdnormcdf` to call the integration.
    See Also
    --------
    mvstdnormcdf : location and scale standardized multivariate normal cdf
    '''
    upper = np.array(upper)
    if lower is None:
        lower = -np.ones(upper.shape) * np.inf
    else:
        lower = np.array(lower)
    cov = np.array(cov)
    stdev = np.sqrt(np.diag(cov))
    lower = (lower - mu) / stdev
    upper = (upper - mu) / stdev
    divrow = np.atleast_2d(stdev)
    corr = cov / divrow / divrow.T
    # v/np.sqrt(np.atleast_2d(np.diag(covv)))/np.sqrt(np.atleast_2d(np.diag(covv))).T

    return _mvstdnormcdf(lower, upper, corr, **kwds)



import numpy as np
from collections import defaultdict
import time
import sparse as sp
import os
from ...backend import STensor
from ...tensor import Graph


def mkdir_self(dir):
    if dir.endswith('.txt'):
        dir_list = dir.split('/')
        dir = '/'.join(dir_list[:-1])
        
    if not os.path.exists(dir):
        os.makedirs(dir)
        
def preprocess_data(stensorList:list, dim):
    print('Dim: ', dim)
    mt_dict, mt_m_dict, concise_a_mt_dict, concise_c_mt_dict = {}, {}, defaultdict(int), defaultdict(int)
    mt_t_dict = {}
    mt_d4_dict = {}  # k_symbols
    m_mt_dict = {} 
    m_mtSize_dict = {} # key: m; value: 对应mt的个数 # for limit size
    
    amt_stensor = stensorList[0]
    cmt_stensor = stensorList[1]
    amt_coo, cmt_coo = amt_stensor._data, cmt_stensor._data
    print(amt_coo.shape)
    a_mt_coo = amt_coo.reshape(shape=(amt_coo.shape[0], -1))  # combine M and T
    c_mt_coo = cmt_coo.reshape(shape=(cmt_coo.shape[0], -1))  # combine M and T
    print(a_mt_coo.shape)
    print(a_mt_coo)
    print(c_mt_coo)

    a_mt_nnz = a_mt_coo.nonzero()
    c_mt_nnz = c_mt_coo.nonzero()

    stt = time.time()
    if dim == 3:
        m_shape_a = amt_stensor.shape[2]
        m_shape_c = cmt_stensor.shape[2]
        for a, mt, elem in zip(a_mt_nnz[0], a_mt_nnz[1], a_mt_coo.data):
            if mt not in mt_dict:
                mt_dict[mt] = len(mt_dict)
            map_m = mt_dict[mt]
            
            if map_m not in mt_m_dict:
                m = mt // m_shape_a
                mt_m_dict[map_m] = m
                if m not in m_mt_dict:
                    m_mt_dict[m] = [map_m]
                    m_mtSize_dict[m] = 1
                else:
                    m_mt_dict[m].append(map_m)
                    m_mtSize_dict[m] += 1
                    
            if map_m not in mt_t_dict:
                mt_t_dict[map_m] = mt % m_shape_a
            key = (a, map_m)
            concise_a_mt_dict[key] += elem # may produce error: unsupported operand type(s) for +=: 'int' and 'str'
        print('time:', time.time() - stt)
        for c, mt, elem in zip(c_mt_nnz[0], c_mt_nnz[1], c_mt_coo.data):
            if mt not in mt_dict:
                mt_dict[mt] = len(mt_dict)
            map_m = mt_dict[mt]
            
            if map_m not in mt_m_dict:
                m = mt // m_shape_c
                mt_m_dict[map_m] = m
                if m not in m_mt_dict:
                    m_mt_dict[m] = [map_m]
                    m_mtSize_dict[m] = 1
                else:
                    m_mt_dict[m].append(map_m)
                    m_mtSize_dict[m] += 1
                    
            if map_m not in mt_t_dict:
                mt_t_dict[map_m] = mt % m_shape_c
            key = (c, map_m)
            concise_c_mt_dict[key] += elem
        print('time:', time.time() - stt)
    elif dim == 4:
        t_shape_a = amt_stensor.shape[2]
        t_shape_c = cmt_stensor.shape[2]
        d4_shape_a = amt_stensor.shape[3]
        d4_shape_c = cmt_stensor.shape[3]
        for a, mt, elem in zip(a_mt_nnz[0], a_mt_nnz[1], a_mt_coo.data):
            if mt not in mt_dict:
                mt_dict[mt] = len(mt_dict)
            map_m = mt_dict[mt]
            
            if map_m not in mt_m_dict:
                m = mt // (t_shape_a * d4_shape_a)
                mt_m_dict[map_m] = m
                if m not in m_mt_dict:
                    m_mt_dict[m] = [map_m]
                    m_mtSize_dict[m] = 1
                else:
                    m_mt_dict[m].append(map_m)
                    m_mtSize_dict[m] += 1
                    
            tmp = mt % (t_shape_a * d4_shape_a)
            if map_m not in mt_t_dict:
                mt_t_dict[map_m] = tmp // d4_shape_a
                
            if map_m not in mt_d4_dict:
                mt_d4_dict[map_m] = tmp % d4_shape_a
            key = (a, map_m)
            concise_a_mt_dict[key] += elem
        print('time:', time.time() - stt)
        for c, mt, elem in zip(c_mt_nnz[0], c_mt_nnz[1], c_mt_coo.data):
            if mt not in mt_dict:
                mt_dict[mt] = len(mt_dict)
            map_m = mt_dict[mt]
            
            if map_m not in mt_m_dict:
                m = mt // (t_shape_c * d4_shape_c)
                mt_m_dict[map_m] = m
                if m not in m_mt_dict:
                    m_mt_dict[m] = [map_m]
                    m_mtSize_dict[m] = 1
                else:
                    m_mt_dict[m].append(map_m)
                    m_mtSize_dict[m] += 1
                    
            tmp = mt % (t_shape_a * d4_shape_a)
            if map_m not in mt_t_dict:
                mt_t_dict[map_m] = tmp // d4_shape_c
                
            if map_m not in mt_d4_dict:
                mt_d4_dict[map_m] = tmp % d4_shape_c
            key = (c, map_m)
            concise_c_mt_dict[key] += elem
        print('time:', time.time() - stt)
    else:
        raise Exception('dim should be 3 or 4!')

    concise_a_mt_coo = sp.COO(concise_a_mt_dict, shape=(a_mt_coo.shape[0], len(mt_dict)))
    concise_c_mt_coo = sp.COO(concise_c_mt_dict, shape=(c_mt_coo.shape[0], len(mt_dict)))
    print(concise_a_mt_coo)
    print(concise_c_mt_coo)

    concise_a_mt_stensor = STensor((concise_a_mt_coo.coords, concise_a_mt_coo.data))
    concise_c_mt_stensor = STensor((concise_c_mt_coo.coords, concise_c_mt_coo.data))

    a_mt_graph = Graph(graph_tensor=concise_a_mt_stensor, bipartite=True, weighted=True)
    c_mt_graph = Graph(graph_tensor=concise_c_mt_stensor, bipartite=True, weighted=True)
    
    print('preprocess Data completed!')
    return [a_mt_graph, c_mt_graph], [mt_m_dict, mt_t_dict, mt_d4_dict], m_mt_dict, m_mtSize_dict

def find_true_m(res, dim, mt_dictList):     
    res_M = []
    res_t = []
    res_d4 = []
    for m in res[0][1]:
        res_M.append(mt_dictList[0][m])
        res_t.append(mt_dictList[1][m])
        if dim == 4:
            res_d4.append(mt_dictList[2][m])
    res_M = list(set(res_M))
    res_t = list(set(res_t))
    if dim == 4:
        res_d4 = list(set(res_d4))
        return res_M, res_t, res_d4
    return res_M, res_t

def get_real_res(res, dim, mt_dictList, outpath):
    if dim == 3:
        res_M, res_t = find_true_m(res, dim, mt_dictList)
    elif dim == 4:
        res_M, res_t, res_d4 = find_true_m(res, dim, mt_dictList)
    else:
        raise Exception('Dim ERR! should be 3 or 4.')
    import copy
    real_res = copy.deepcopy(res)
    real_res[0][1] = res_M
    real_res[0].append(res_t)
    if dim == 4:
        real_res[0].append(res_d4)
    print('Find block:')
    if dim == 3:
        print(len(real_res[0][0]), len(real_res[0][1]), len(real_res[0][2]), len(real_res[0][3]))
    else:
        print(len(real_res[0][0]), len(real_res[0][1]), len(real_res[0][2]), len(real_res[0][3]), len(real_res[0][4]))
    if outpath != '':
        saveres2txt(real_res, outpath)
    return real_res

def saveres2npy(res, outpath):
    mkdir_self(outpath)
    for i in range(len(res[0])):
        res[0][i] = list(res[0][i])
    nres = []
    resAddr = outpath+'res.npy'
    if os.path.exists(resAddr):
        nres = np.load(resAddr, allow_pickle=True)
        nres = nres.tolist()
    nres.extend([res])
    np.save(resAddr, nres)
    print('res save completed!')

def saveres2txt(res, outpath):
    mkdir_self(outpath)
    for i in range(len(res[0])):
        res[0][i] = list(res[0][i])
    with open(outpath, "a") as f:
        f.write('[')
        f.write('\n')
        for i in range(len(res[0])):
            for j in range(len(res[0][i])):
                f.write(str(res[0][i][j]))
                f.write('\t')
            f.write('\n')
        f.write(str(res[1]))
        f.write('\n')
        f.write(']')
        f.write('\n')
    print('save res.txt success!')

def loadtxt2res(outpath, dim, data_type=int):
    res = []
    tmp_res = []
    with open(outpath, "r") as f:
        contents =f.readlines()
        for line in contents:
            if line.startswith('['):
                tmp_res = []
                continue
            if line.startswith(']'):
                res.append(tmp_res)
                continue
            acc_list = line.strip().split("\t")
            if len(tmp_res) < dim + 1:
                acc_list = list(map(data_type, acc_list))
                tmp_res.append(acc_list)
            else:
                score = float(acc_list[0])
                tmp_res = [tmp_res, score]
    print('load res.txt success!')
    return res
        
def get_zero_matrix(idx, size):
    row = np.array(list(range(size)))
    col = np.array(list(range(size)))
    data = np.array([1]*size)
    row = np.delete(row, idx)
    col = np.delete(col, idx)
    data = np.delete(data, idx)
    return sp.COO((data, (row, col)), shape=(size, size))
    
    
    
    



from numpy import random

from .SinglePassPCA import generateGH_by_multiply, generateQB, computeSVD
from .util import calDensity, findSuspWins, filterEigenvec, getKeys
from spartan.tensor import TensorData, DTensor
from .._model import DMmodel
from spartan.util.basicutil import StringMapper
from spartan.util.drawutil import drawEigenPulse


class EigenPulse(DMmodel):
    def __init__(self, stream_tensor, **param_dict):
        self.stream_tensor = stream_tensor
        self.window = param_dict['window']
        self.stride = param_dict['stride']
        self.l = param_dict['l']
        self.b = param_dict['b']
        self.ts_colidx = param_dict['ts_idx']
        self.item_colidx = param_dict['item_idx']
        self.stringMapper = StringMapper()
        self.mappers = {self.item_colidx: self.stringMapper}

    def anomaly_detection(self):
        return self.run()

    def run(self):
        densities, submats = [], []
        winidx = 0
        while True:
            try:
                tensorlist = self.stream_tensor.fetch_slide_window(self.window, self.stride, self.ts_colidx)
                tensorlist[self.item_colidx] = \
                    tensorlist[[self.item_colidx, self.ts_colidx]].apply(tuple, axis=1)
                del tensorlist[self.ts_colidx]

                td = TensorData(tensorlist)
                stensor = td.toSTensor(self.stream_tensor.hasvalue, self.mappers)
 
                n = stensor.shape[1]
                
                Omg = DTensor.from_numpy(x=random.randn(n, self.l))
                G, H = generateGH_by_multiply(stensor, Omg)
                Q, B = generateQB(G, H, Omg, self.l, self.b)
                u1, s1, v1 = computeSVD(Q, B)
                u1 = u1.T
                submat, rowids, colids = filterEigenvec(stensor, u1, v1)
                submats.append((submat, rowids, colids))

                density = calDensity(submat)
                print(f"density of dense submatrix in window {winidx} is {density}")
                densities.append(density)
                winidx += 1
            except:
                break
        
        susp_wins = findSuspWins(densities)
        print('suspicious windows:', susp_wins)
        res = []
        for susp_win in susp_wins:
            submat, rowids, colids = submats[susp_win]
            items, tss = getKeys(colids, self.mappers[1].strdict)
            res.append((susp_win, rowids, items, tss, densities[susp_win]))
        for block in res:
            print('suspicious window:', block[0])
            print('user list of dense block:', block[1])
            print('item list of dense block:', block[2])
            print('time list of dense block:', block[3])
            print('density of dense block:', block[4])
        return res, densities



#!/usr/bin/python2.7
# -*- coding=utf-8 -*-

#  Project: eaglemine
#      loader.py
#      Created by @wenchieh  on <11/25/2017>


__author__ = 'wenchieh'


# third-party lib
import numpy as np


class Loader(object):
    def _get_sep_of_file_(self, infn):
        '''
        return the separator of the line.
        :param infn: input file
        '''
        sep = None
        with open(infn, 'r') as fp:
            for line in fp:
                if (line.startswith("%") or line.startswith("#")): continue;
                line = line.strip()
                if (" " in line): sep = " "
                if ("," in line): sep = ","
                if (";" in line): sep = ';'
                if ("\t" in line): sep = "\t"
                break
            fp.close()
        return sep

    def load_edgelist(self, infn_edgelist, comments='#', dtype=int,
                      usecols=(0, 1), idstartzero=True):
        '''
        load edge-list file (node type: int; start index: 0)
        :param infn_edgelist: edgelist file name
        :param comments: comments for the file (at the first line) (default: #)
        :param dtype: all_data type (default: int)
        :param usecols: select specific columns in file (default: [0,1])
        :param idstartzero: whether the id start from zero or not.
        :return: numpy array
        '''
        sep = self._get_sep_of_file_(infn_edgelist)
        data = []

        offset = 0
        if idstartzero != True:
            offset = -1
        with open(infn_edgelist) as fp:
            for line in fp:
                if line.startswith(comments):
                    continue
                arr = line.strip().split(sep)
                elem = []
                for idx in usecols:
                    elem.append(dtype(arr[idx]) + offset)
                data.append(np.array(elem))
            fp.close()

        return np.array(data)

    def load_edgelist2dict(self, infn_edgelist, weighted=False, comments='#',
                           dtype=int, uvindex=(0, 1), idstartzero=True):
        sep = self._get_sep_of_file_(infn_edgelist)
        edges_set = set()
        u2v, v2u = dict(), dict()

        offset = 0
        if idstartzero != True:
            offset = -1
        with open(infn_edgelist) as fp:
            for line in fp:
                if line.startswith(comments):
                    continue
                tokens = line.strip().split(sep)
                u = dtype(tokens[uvindex[0]]) + offset
                v = dtype(tokens[uvindex[1]]) + offset
                if weighted:
                    if u not in u2v:
                        u2v[u] = list()
                    if v not in v2u:
                        v2u[v] = list()
                    u2v[u].append(v)
                    v2u[v].append(u)
                else:
                    if (u, v) not in edges_set:
                        if u not in u2v:
                            u2v[u] = list()
                        if v not in v2u:
                            v2u[v] = list()
                        u2v[u].append(v)
                        v2u[v].append(u)
                        edges_set.add((u, v))

        return u2v, v2u

    def load_adjacency(self, infn_adjlist):
        '''
        load adjacency list [file_adjacencylist] to dict
        :param fn_adjacencylist: input file of adjacency list
        :return: adjacency list dictionary
        '''
        adj_list = dict()

        sep = self._get_sep_of_file_(infn_adjlist)
        with open(infn_adjlist, 'r') as fp:
            for line in fp:
                arr = line.split(sep)
                adj_list[int(arr[0])] = list(map(int, arr[1].split(sep)))
            fp.close()

        return adj_list

    def load_features(self, infn, val_type):
        sep = self._get_sep_of_file_(infn)

        with open(infn, 'r') as fp:
            line = fp.readline().strip()
            m, n = map(int, line[1:].split(sep)[:2])
            fp.close()
        features = np.loadtxt(infn, val_type, delimiter=sep)

        return m, n, features

    def load_degvsfeats(self, infn, val_type, usecols=(0, 1)):
        sep = self._get_sep_of_file_(infn)

        with open(infn, 'r') as fp:
            line = fp.readline()
            splited_line = line[1:].strip().split(sep)
            m, n, mod = int()
            m, n, mod = map(int, line[1:].strip().split(sep))
            usecols = np.unique(usecols)
            degs = []   # np.zeros(m + n, int)
            vals = []   # np.zeros(m + n, val_type)

            for ln in range(m + n):
                arr = fp.readline().strip().split(sep)
                if usecols is None or 0 in usecols:
                    degs.append(int(arr[0]))

                if usecols is None or len(usecols) == mod:
                    rec_val = val_type(arr)
                else:
                    feat_vals = []
                    for k in usecols:
                        if k > 0 and k < mod:
                            feat_vals.append(val_type(arr[k]))
                    if len(feat_vals) == 1:
                        rec_val = feat_vals[0]
                    else:
                        rec_val = feat_vals
                vals.append(rec_val)
            fp.close()

        return m, n, degs, vals

    def load_histogram(self, infn):
        return self.load_multi_histogram(infn)

    def load_multi_histogram(self, infn):
        shape, ticks_vec = [], []
        hist_arr = []

        sep = self._get_sep_of_file_(infn)
        with open(infn, 'r') as ofp:
            shape = list(map(int, ofp.readline().strip()[1:].split(sep)))

            for mod in range(len(shape)):
                ticks = list(map(float, ofp.readline()[1:].strip().split(sep)))
                ticks_vec.append(ticks)

            for line in ofp.readlines():
                toks = line.strip().split(sep)
                hist_arr.append((list(map(int, toks))))
            ofp.close()

        return np.array(shape, int), ticks_vec, np.array(hist_arr, int)

    def load_hcubepos2pts(self, pts2pos_infn, comment='#'):
        pos2pts = dict()
        sep = self._get_sep_of_file_(pts2pos_infn)
        with open(pts2pos_infn, 'r') as ifp:
            index = 0
            for line in ifp:
                if line.startswith(comment):
                    continue
                pos = tuple(map(int, line.strip().split(sep)))
                if pos not in pos2pts:
                    pos2pts[pos] = []
                pos2pts[pos].append(index)
                index += 1
            ifp.close()
        return pos2pts

    def load_pt2pos(self, pts2pos_infn, comment='#'):
        sep = self._get_sep_of_file_(pts2pos_infn)
        ptidx_pos = list()
        with open(pts2pos_infn, 'r') as ifp:
            for line in ifp:
                if line.startswith(comment):
                    continue
                toks = list(map(int, line.strip().split(sep)))
                ptidx_pos.append(toks)
            ifp.close()
        return np.asarray(ptidx_pos)

    def load_describes_parms(sel, describe_infn, describe_cls, mode=2, sep=';'):
        params = list()
        with open(describe_infn, 'r') as ifp:
            for line in ifp:
                line = line.strip()
                splits = line.find(sep)
                desc = describe_cls(mode)
                desc.load(line[splits+1:])
                if desc.is_mix:
                    for k in range(desc.n_components):
                        cov = desc.paras['covs'][k]
                        kparm = list(desc.paras['mus'][k])
                        for m in range(desc.ndim):
                            kparm.extend(cov[m, m:])
                        params.append(kparm)
                else:
                    params.append(desc.compact_parm())
            ifp.close()

        return params